"""
    SoftLayer.transports
    ~~~~~~~~~~~~~~~~~~~~
    XML-RPC transport layer that uses the requests library.

    :license: MIT, see LICENSE for more details.
"""
# Required imports to not break existing code.


from .debug import DebugTransport
from .fixture import FixtureTransport
from .rest import RestTransport
from .timing import TimingTransport
from .transport import Request
from .transport import SoftLayerListResult as SoftLayerListResult
from .xmlrpc import XmlRpcTransport

# transports.Request does have a lot of instance attributes. :(
# pylint: disable=too-many-instance-attributes

__all__ = [
    'Request',
    'XmlRpcTransport',
    'RestTransport',
    'TimingTransport',
    'DebugTransport',
    'FixtureTransport',
    'SoftLayerListResult'
]



createCancelServerTicket = {'id': 1234, 'title': 'Server Cancellation Request'}
getObject = {
    "administrativeBypassFlag": "",
    "customerManagedFlag": False,
    "billingItem": {
        "id": 21370815
    },
    "id": 3130,
    "primaryIpAddress": "192.155.239.146",
    "datacenter": {
        "id": 265592,
        "longName": "Amsterdam 1",
        "name": "ams01",
        "statusId": 2
    },
    "networkVlan": {
        "accountId": 307608,
        "id": 371028,
        "primarySubnetId": 536252,
        "name": 'testvlan',
        "vlanNumber": 1489,
        "firewallInterfaces": [
            {
                "id": 6254,
                "name": "inside",
                "firewallContextAccessControlLists": [
                    {
                        "direction": "out",
                        "firewallContextInterfaceId": 6257,
                        "id": 3143
                    }
                ]
            },
            {
                "id": 6256,
                "name": "outside",
                "firewallContextAccessControlLists": [
                    {
                        "direction": "out",
                        "firewallContextInterfaceId": 6257,
                        "id": 3143
                    },
                    {
                        "direction": "in",
                        "firewallContextInterfaceId": 6256,
                        "id": 3142
                    }
                ]
            }
        ]
    },
    "firewallType": "fortigate-security-appliance-10gb",
    "managementCredentials": {
        "createDate": "2022-05-17T13:59:17-06:00",
        "id": 74604882,
        "modifyDate": "2022-05-17T13:59:17-06:00",
        "password": "test1234",
        "port": 23,
        "softwareId": 67804284,
        "username": "myusername"
    },
    "networkGateway": {
        "accountId": 307608,
        "groupNumber": 1,
        "id": 615448,
        "name": "testFirewall",
        "networkSpace": "BOTH",
        "privateIpAddressId": 188996652,
        "privateVlanId": 3228724,
        "publicIpAddressId": 188996794,
        "publicIpv6AddressId": 188996808,
        "publicVlanId": 3228726,
        "statusId": 1,
        "insideVlans": [],
        "members": [
            {
                "hardwareId": 3222842,
                "id": 687820,
                "networkGatewayId": 615448,
                "priority": 254,
                "networkGateway": None
            }
        ],
        "privateIpAddress": {
            "id": 188996652,
            "ipAddress": "10.37.115.70",
            "isBroadcast": False,
            "isGateway": False,
            "isNetwork": False,
            "isReserved": True,
            "subnetId": 2552734,
            "subnet": {
                "broadcastAddress": "10.37.115.127",
                "cidr": 26,
                "gateway": "10.37.115.65",
                "id": 2552734,
                "isCustomerOwned": False,
                "isCustomerRoutable": False,
                "modifyDate": "2022-05-17T13:59:16-06:00",
                "netmask": "255.255.255.192",
                "networkIdentifier": "10.37.115.64",
                "networkVlanId": 3228724,
                "sortOrder": "1",
                "subnetType": "ADDITIONAL_PRIMARY",
                "totalIpAddresses": "64",
                "usableIpAddressCount": "61",
                "version": 4
            }
        },
        "privateVlan": {
            "accountId": 307608,
            "fullyQualifiedName": "dal13.bcr03.1330",
            "id": 3228724,
            "modifyDate": "2022-05-17T14:01:14-06:00",
            "primarySubnetId": 2625456,
            "vlanNumber": 1330
        },
        "publicIpAddress": {
            "id": 188996794,
            "ipAddress": "67.228.206.245",
            "isBroadcast": False,
            "isGateway": False,
            "isNetwork": False,
            "isReserved": True,
            "subnetId": 66444,
            "subnet": {
                "broadcastAddress": "67.228.206.247",
                "cidr": 29,
                "gateway": "67.228.206.241",
                "id": 66444,
                "isCustomerOwned": False,
                "isCustomerRoutable": False,
                "modifyDate": "2022-05-17T13:59:16-06:00",
                "netmask": "255.255.255.248",
                "networkIdentifier": "67.228.206.240",
                "networkVlanId": 3228726,
                "sortOrder": "1",
                "subnetType": "ADDITIONAL_PRIMARY",
                "totalIpAddresses": "8",
                "usableIpAddressCount": "5",
                "version": 4
            }
        },
        "publicIpv6Address": {
            "id": 188996808,
            "ipAddress": "2607:f0d0:2703:0039:0000:0000:0000:0004",
            "isBroadcast": False,
            "isGateway": False,
            "isNetwork": False,
            "isReserved": True,
            "subnetId": 2547678,
            "subnet": {
                "broadcastAddress": "",
                "cidr": 64,
                "gateway": "2607:f0d0:2703:0039:0000:0000:0000:0001",
                "id": 2547678,
                "isCustomerOwned": False,
                "isCustomerRoutable": False,
                "modifyDate": "2022-05-17T13:59:16-06:00",
                "netmask": "ffff:ffff:ffff:ffff:0000:0000:0000:0000",
                "networkIdentifier": "2607:f0d0:2703:0039:0000:0000:0000:0000",
                "networkVlanId": 3228726,
                "sortOrder": "4",
                "subnetType": "PRIMARY_6",
                "totalIpAddresses": "18446744073709551616",
                "usableIpAddressCount": "18446744073709551614",
                "version": 6
            }
        },
        "publicVlan": {
            "accountId": 307608,
            "fullyQualifiedName": "dal13.fcr03.1255",
            "id": 3228726,
            "modifyDate": "2022-05-17T14:00:42-06:00",
            "primarySubnetId": 2623338,
            "vlanNumber": 1255
        },
        "status": {
            "description": "Gateway is ready to accept commands and actions",
            "id": 1,
            "keyName": "ACTIVE",
            "name": "Active"
        }
    },
    "rules": [
        {'destinationIpAddress': 'any on server',
         'protocol': 'tcp',
         'orderValue': 1,
         'destinationIpSubnetMask': '255.255.255.255',
         'destinationPortRangeStart': 80,
         'sourceIpSubnetMask': '0.0.0.0',
         'destinationPortRangeEnd': 80,
         'version': 4,
         'action': 'permit',
         'sourceIpAddress': '0.0.0.0'
         },
        {
            'destinationIpAddress': 'any on server',
            'protocol': 'tcp',
            'orderValue': 2,
            'destinationIpSubnetMask': '255.255.255.255',
            'destinationPortRangeStart': 1,
            'sourceIpSubnetMask': '255.255.255.255',
            'destinationPortRangeEnd': 65535,
            'version': 4,
            'action': 'permit',
            'sourceIpAddress': '193.212.1.10'
        },
        {
            'destinationIpAddress': 'any on server',
            'protocol': 'tcp',
            'orderValue': 3,
            'destinationIpSubnetMask': '255.255.255.255',
            'destinationPortRangeStart': 80,
            'sourceIpSubnetMask': '0.0.0.0',
            'destinationPortRangeEnd': 800,
            'version': 4,
            'action': 'permit',
            'sourceIpAddress': '0.0.0.0'
        }
    ],
    "metricTrackingObject": {
        "id": 147258369,
        "resourceTableId": 17438,
        "startDate": "2022-05-17T14:01:48-06:00",
        "type": {
            "keyname": "FIREWALL_CONTEXT",
            "name": "Firewall Module Context"
        }
    }

}

getRules = [
    {
        'destinationIpAddress': 'any on server',
        'protocol': 'tcp',
        'orderValue': 1,
        'destinationIpSubnetMask': '255.255.255.255',
        'destinationPortRangeStart': 80,
        'sourceIpSubnetMask': '0.0.0.0',
        'destinationPortRangeEnd': 80,
        'version': 4,
        'action': 'permit',
        'sourceIpAddress': '0.0.0.0'
    },
    {
        'destinationIpAddress': 'any on server',
        'protocol': 'tmp',
        'orderValue': 2,
        'destinationIpSubnetMask': '255.255.255.255',
        'destinationPortRangeStart': 1,
        'sourceIpSubnetMask': '255.255.255.255',
        'destinationPortRangeEnd': 65535,
        'version': 4,
        'action': 'permit',
        'sourceIpAddress': '193.212.1.10'
    },
    {
        'destinationIpAddress': 'any on server',
        'protocol': 'tcp',
        'orderValue': 3,
        'destinationIpSubnetMask': '255.255.255.255',
        'destinationPortRangeStart': 80,
        'sourceIpSubnetMask': '0.0.0.0',
        'destinationPortRangeEnd': 800,
        'version': 4,
        'action': 'permit',
        'sourceIpAddress': '0.0.0.0'
    }
]
edit = True



createCancelServerTicket = {'id': 1234, 'title': 'Server Cancellation Request'}
getObject = {
    "accountId": 1234,
    "serviceProviderResourceId": "CS123456",
    "assignedUserId": 12345,
    "createDate": "2013-08-01T14:14:04-07:00",
    "id": 100,
    "lastEditDate": "2013-08-01T14:16:47-07:00",
    "lastEditType": "AUTO",
    "modifyDate": "2013-08-01T14:16:47-07:00",
    "status": {
        "id": 1002,
        "name": "Closed"
    },
    "statusId": 1002,
    "title": "Cloud Instance Cancellation - 08/01/13",
    'updateCount': 3,
    'updates': [
        {'entry': 'a bot says something'},
        {'entry': 'user says something',
         'editor': {'firstName': 'John', 'lastName': 'Smith'}},
        {'entry': 'employee says something',
         'editor': {'displayName': 'emp1'}},

    ]
}

createStandardTicket = {
    "assignedUserId": 12345,
    "serviceProviderResourceId": "CS123456",
    "id": 100,
    "contents": "body",
    "subjectId": 1004,
    "title": "Cloud Instance Cancellation - 08/01/13"
}
edit = True
addUpdate = {}

gatList = getObject

addAttachedHardware = {
    "id": 123,
    "createDate": "2013-08-01T14:14:04-07:00",
    "hardwareId": 1,
    "ticketId": 100
}

addAttachedVirtualGuest = {
    "id": 123,
    "createDate": "2013-08-01T14:14:04-07:00",
    "virtualGuestId": 1,
    "ticketId": 100
}

addAttachedFile = {
    "id": 123,
    "fileName": "a_file_name",
    "fileSize": "100",
    "ticketId": 100,
    "updateId": 200,
    "createDate": "2013-08-01T14:14:04-07:00",
    "modifyDate": "2013-08-01T14:14:04-07:00",
    "uploaderType": "USER",
    "uploaderId": "300"
}

removeAttachedHardware = True
removeAttachedVirtualGuest = True



"""List dedicated servers."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@helpers.multi_option('--tag', help='Filter by tags')
@click.option('--sortby', help='Column to sort by',
              default='Name',
              show_default=True)
@click.option('--datacenter', '-d', help='Filter by datacenter shortname')
@click.option('--name', '-H', help='Filter by host portion of the FQDN')
@click.option('--order', help='Filter by ID of the order which purchased this dedicated host', type=click.INT)
@click.option('--owner', help='Filter by owner of the dedicated host')
@environment.pass_env
def cli(env, sortby, datacenter, name, tag, order, owner):
    """List dedicated host."""
    mgr = SoftLayer.DedicatedHostManager(env.client)
    dedicated_hosts = mgr.list_instances(datacenter=datacenter,
                                         hostname=name,
                                         tags=tag,
                                         order=order,
                                         owner=owner)

    table = formatting.Table(["Id", "Name", "Datacenter", "Router", "CPU (allocated/total)",
                              "Memory (allocated/total)", "Disk (allocated/total)", "Guests"])
    table.align['Name'] = 'l'
    table.sortby = sortby

    if len(dedicated_hosts) != 0:
        for host in dedicated_hosts:
            cpu_allocated = host.get('allocationStatus').get('cpuAllocated')
            cpu_total = host.get('allocationStatus').get('cpuCount')
            memory_allocated = host.get('allocationStatus').get('memoryAllocated')
            memory_total = host.get('allocationStatus').get('memoryCapacity')
            disk_allocated = host.get('allocationStatus').get('diskAllocated')
            disk_total = host.get('allocationStatus').get('diskCapacity')
            table.add_row([
                host.get('id'),
                host.get('name'),
                host.get('datacenter').get('name'),
                host.get('backendRouter').get('hostname'),
                f"{cpu_allocated}/{cpu_total}",
                f"{memory_allocated}/{memory_total}",
                f"{disk_allocated}/{disk_total}",
                host.get('guestCount')
            ])

        env.fout(table)
    else:
        click.secho("No dedicated hosts are found.")



"""Gets some details about a specific billing item."""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Gets detailed information about a billing item."""
    manager = AccountManager(env.client)
    item = manager.get_item_detail(identifier)
    env.fout(item_table(item))


def item_table(item):
    """Formats a table for billing items"""

    date_format = '%Y-%m-%d'
    table = formatting.Table(["Key", "Value"], title=f"{item.get('description', 'Billing Item')}")
    table.add_row(['createDate', utils.clean_time(item.get('createDate'), date_format, date_format)])
    table.add_row(['cycleStartDate', utils.clean_time(item.get('cycleStartDate'), date_format, date_format)])
    table.add_row(['cancellationDate', utils.clean_time(item.get('cancellationDate'), date_format, date_format)])
    table.add_row(['description', item.get('description')])
    table.align = 'l'
    fqdn = f"{item.get('hostName', '')}.{item.get('domainName', '')}"
    if fqdn != ".":
        table.add_row(['FQDN', fqdn])

    if item.get('hourlyFlag', False):
        table.add_row(['hourlyRecurringFee', item.get('hourlyRecurringFee')])
        table.add_row(['hoursUsed', item.get('hoursUsed')])
        table.add_row(['currentHourlyCharge', item.get('currentHourlyCharge')])
    else:
        table.add_row(['recurringFee', item.get('recurringFee')])

    ordered_by = "IBM"
    user = utils.lookup(item, 'orderItem', 'order', 'userRecord')
    if user:
        ordered_by = f"{user.get('displayName')} ({utils.lookup(user, 'userStatus', 'name')})"
    table.add_row(['Ordered By', ordered_by])
    table.add_row(['Notes', item.get('notes')])
    table.add_row(['Location', utils.lookup(item, 'location', 'name')])
    if item.get('children'):
        for child in item.get('children'):
            table.add_row([child.get('categoryCode'), child.get('description')])

    return table



""" Credential limits for this IBM Cloud Object Storage account."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Credential limits for this IBM Cloud Object Storage account."""

    mgr = SoftLayer.ObjectStorageManager(env.client)
    storage_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'Object Storage')
    credential_limit = mgr.limit_credential(storage_id)
    table = formatting.Table(['limit'])
    table.add_row([
        credential_limit,
    ])

    env.fout(table)



"""List SSH keys."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--sortby',
              help='Column to sort by',
              type=click.Choice(['id',
                                 'label',
                                 'fingerprint',
                                 'notes']))
@environment.pass_env
def cli(env, sortby):
    """List SSH keys."""

    mgr = SoftLayer.SshKeyManager(env.client)
    keys = mgr.list_keys()

    table = formatting.Table(['id', 'label', 'fingerprint', 'notes'])
    table.sortby = sortby

    for key in keys:
        table.add_row([key['id'],
                       key.get('label'),
                       key.get('fingerprint'),
                       key.get('notes', '-')])

    env.fout(table)



"""List cloud object storage volumes."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@environment.pass_env
def cli(env):
    """List cloud block storage."""
    block_manager = SoftLayer.BlockStorageManager(env.client)

    storages = block_manager.get_cloud_list()

    table = formatting.Table(['Id',
                              'Account name',
                              'Description',
                              'Create Date',
                              'Type'])
    for storage in storages:
        table.add_row([storage.get('id'),
                       storage.get('username'),
                       storage['storageType']['description'],
                       storage['billingItem']['createDate'],
                       storage['storageType']['keyName']])

    env.fout(table)



"""Get bandwidth pool details."""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer import AccountManager
from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get bandwidth pool details."""

    manager = AccountManager(env.client)
    bandwidths = manager.getBandwidthDetail(identifier)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Id', bandwidths['id']])
    table.add_row(['Name', bandwidths['name']])
    table.add_row(['Create Date', utils.clean_time(bandwidths.get('createDate'), '%Y-%m-%d')])
    end_date = utils.clean_time(bandwidths.get('endDate'))
    if end_date == '':
        end_date = formatting.blank()
        table.add_row(['End Date', end_date])
    else:
        table.add_row(['End Date', utils.clean_time(bandwidths.get('endDate'))])
    current = f"{utils.lookup(bandwidths, 'billingCyclePublicBandwidthUsage', 'amountOut')} GB"
    if current is None:
        current = '-'
    table.add_row(['Current Usage', current])
    projected = f"{bandwidths.get('projectedPublicBandwidthUsage', 0)} GB"
    if projected is None:
        projected = '-'
    table.add_row(['Projected  Usage', projected])
    inbound = f"{bandwidths.get('inboundPublicBandwidthUsage', 0)} GB"
    if inbound is None:
        inbound = '-'
    table.add_row(['Inbound   Usage', inbound])
    if bandwidths['hardware'] != []:
        table.add_row(['hardware', *(_bw_table(bandwidths['hardware']))])
    else:
        table.add_row(['hardware', 'Not Found'])

    if bandwidths['virtualGuests'] != []:
        table.add_row(['virtualGuests', *(_virtual_table(bandwidths['virtualGuests']))])
    else:
        table.add_row(['virtualGuests', 'Not Found'])

    if bandwidths['bareMetalInstances'] != []:
        table.add_row(['Netscaler', *(_bw_table(bandwidths['bareMetalInstances']))])
    else:
        table.add_row(['Netscaler', 'Not Found'])

    env.fout(table)


def _bw_table(bw_data):
    """Generates a bandwidth useage table"""
    table_data = formatting.Table(['Id', 'HostName', "IP Address", 'Amount', "Current Usage"])
    for bw_point in bw_data:
        amount = f"{utils.lookup(bw_point, 'bandwidthAllotmentDetail', 'allocation', 'amount')} GB"
        current = f"{bw_point.get('outboundBandwidthUsage', 0)} GB"
        ip_address = bw_point.get('primaryIpAddress')
        if ip_address is None:
            ip_address = '-'
        table_data.add_row([bw_point['id'], bw_point['fullyQualifiedDomainName'], ip_address, amount, current])
    return [table_data]


def _virtual_table(bw_data):
    """Generates a virtual bandwidth usage table"""
    table_data = formatting.Table(['Id', 'HostName', "IP Address", 'Amount', "Current Usage"])
    for bw_point in bw_data:
        amount = f"{utils.lookup(bw_point, 'bandwidthAllotmentDetail', 'allocation', 'amount')} GB"
        current = f"{bw_point.get('outboundBandwidthUsage', 0)} GB"
        ip_address = bw_point.get('primaryIpAddress')
        if ip_address is None:
            ip_address = '-'
        table_data.add_row([bw_point['id'], bw_point['fullyQualifiedDomainName'], ip_address, amount, current])
    return [table_data]



"""List all options for ordering a file storage."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI.command import SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting

PACKAGE_STORAGE = 759


@click.command(cls=SLCommand)
@click.argument('location', required=False)
@click.option('--prices', '-p', is_flag=True,
              help='Use --prices to list the server item prices, and to list the Item Prices by location,'
                   'add it to the --prices option using location short name, e.g. --prices dal13')
@environment.pass_env
def cli(env, prices, location=None):
    """List all options for ordering a block storage

    Example::
            slcli file volume-options
            slcli file volume-options --prices dal13
    """

    order_manager = SoftLayer.OrderingManager(env.client)
    items = order_manager.get_items(PACKAGE_STORAGE)
    datacenters = order_manager.get_regions(PACKAGE_STORAGE, location)

    tables = []
    network = SoftLayer.NetworkManager(env.client)

    pods = network.get_closed_pods()

    if datacenters != []:
        datacenter_table = formatting.Table(['Id', 'Description', 'KeyName'], title='Datacenter')

        for datacenter in datacenters:
            closure = []
            for pod in pods:
                if datacenter['location']['location']['name'] in str(pod['name']):
                    closure.append(pod['name'])

            notes = '-'
            if len(closure) > 0:
                notes = 'closed soon: %s' % (', '.join(closure))
            datacenter_table.add_row([datacenter['location']['locationId'],
                                      datacenter.get('description'),
                                      datacenter['keyname'], notes])
        tables.append(datacenter_table)
    else:
        raise exceptions.CLIAbort('Location does not exit.')

    tables.append(_file_ios_get_table(items, prices))
    tables.append(_file_storage_table(items, prices))
    tables.append(_file_snapshot_get_table(items, prices))
    env.fout(tables)


def _file_ios_get_table(items, prices):
    if prices:
        table = formatting.Table(['Id', 'Description', 'KeyName', 'Prices'], title='IOPS')
        for item in items:
            if item['itemCategory']['categoryCode'] == 'storage_tier_level':
                table.add_row([item.get('id'), item.get('description'),
                               item.get('keyName'), item['prices'][0]['recurringFee']])
    else:
        table = formatting.Table(['Id', 'Description', 'KeyName'], title='IOPS')
        for item in items:
            if item['itemCategory']['categoryCode'] == 'storage_tier_level':
                table.add_row([item.get('id'), item.get('description'),
                               item.get('keyName')])
    table.sortby = 'Id'
    table.align = 'l'
    return table


def _file_storage_table(items, prices):
    if prices:
        table = formatting.Table(['Id', 'Description', 'KeyName', 'Capacity Minimum', 'Prices'], title='Storage')
        for item in items:
            if item['itemCategory']['categoryCode'] == 'performance_storage_space':
                table.add_row([item.get('id'), item.get('description'),
                               item.get('keyName'), item.get('capacityMinimum') or '-',
                               item['prices'][0]['recurringFee']])
    else:
        table = formatting.Table(['Id', 'Description', 'KeyName', 'Capacity Minimum'], title='Storage')
        for item in items:
            if item['itemCategory']['categoryCode'] == 'performance_storage_space':
                table.add_row([item.get('id'), item.get('description'),
                               item.get('keyName'), item.get('capacityMinimum') or '-', ])
    table.sortby = 'Id'
    table.align = 'l'
    return table


def _file_snapshot_get_table(items, prices):
    if prices:
        table = formatting.Table(['Id', 'Description', 'KeyName', 'Prices'], title='Snapshot')
        for item in items:
            if item['itemCategory']['categoryCode'] == 'storage_snapshot_space':
                table.add_row([item.get('id'), item.get('description'),
                               item.get('keyName'), item['prices'][0]['recurringFee']])
    else:
        table = formatting.Table(['Id', 'Description', 'KeyName'], title='Snapshot')
        for item in items:
            if item['itemCategory']['categoryCode'] == 'storage_snapshot_space':
                table.add_row([item.get('id'), item.get('description'),
                               item.get('keyName')])
    table.sortby = 'Id'
    table.align = 'l'
    return table



"""
    SoftLayer.storage
    ~~~~~~~~~~~~~~~~~
    Network Storage Manager

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer import exceptions
from SoftLayer.managers import storage_utils
from SoftLayer import utils


# pylint: disable=too-many-public-methods


class StorageManager(utils.IdentifierMixin, object):
    """"Base class for File and Block storage managers

    Any shared code between File and Block should ideally go here.

    :param SoftLayer.API.BaseClient client: the client instance
    """

    def __init__(self, client):
        self.configuration = {}
        self.client = client
        self.resolvers = [self._get_ids_from_username]

    def _get_ids_from_username(self, username):  # pylint: disable=unused-argument
        """Should only be actually called from the block/file manager"""
        return []

    def get_volume_count_limits(self):
        """Returns a list of block volume count limit.

        :return: Returns a list of block volume count limit.
        """
        return self.client.call('Network_Storage', 'getVolumeCountLimits')

    def get_volume_details(self, volume_id, **kwargs):
        """Returns details about the specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns details about the specified volume.
        """

        if 'mask' not in kwargs:
            items = [
                'id',
                'username',
                'password',
                'capacityGb',
                'snapshotCapacityGb',
                'parentVolume.snapshotSizeBytes',
                'storageType.keyName',
                'serviceResource.datacenter[name]',
                'serviceResourceBackendIpAddress',
                'storageTierLevel',
                'provisionedIops',
                'lunId',
                'originalVolumeName',
                'originalSnapshotName',
                'originalVolumeSize',
                'activeTransactionCount',
                'activeTransactions.transactionStatus[friendlyName]',
                'replicationPartnerCount',
                'replicationStatus',
                'replicationPartners[id,username,'
                'serviceResourceBackendIpAddress,'
                'serviceResource[datacenter[name]],'
                'replicationSchedule[type[keyname]]]',
                'notes',
            ]
            kwargs['mask'] = ','.join(items)
        return self.client.call('Network_Storage', 'getObject', id=volume_id, **kwargs)

    def get_volume_access_list(self, volume_id, **kwargs):
        """Returns a list of authorized hosts for a specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns a list of authorized hosts for a specified volume.
        """
        if 'mask' not in kwargs:
            items = [
                'id',
                'allowedVirtualGuests[allowedHost[credential, sourceSubnet]]',
                'allowedHardware[allowedHost[credential]]',
                'allowedSubnets[allowedHost[credential]]',
                'allowedIpAddresses[allowedHost[credential]]',
            ]
            kwargs['mask'] = ','.join(items)
        return self.client.call('Network_Storage', 'getObject', id=volume_id, **kwargs)

    def get_volume_snapshot_list(self, volume_id, **kwargs):
        """Returns a list of snapshots for the specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns a list of snapshots for the specified volume.
        """
        if 'mask' not in kwargs:
            items = [
                'id', 'notes', 'snapshotSizeBytes', 'storageType[keyName]',
                'snapshotCreationTimestamp', 'intervalSchedule',
                'hourlySchedule', 'dailySchedule', 'weeklySchedule'
            ]

            kwargs['mask'] = ','.join(items)

        return self.client.call('Network_Storage', 'getSnapshots', id=volume_id, **kwargs)

    def set_volume_snapshot_notification(self, volume_id, enable):
        """Enables/Disables snapshot space usage threshold warning for a given volume.

        :param volume_id: ID of volume.
        :param enable: Enable/Disable flag for snapshot warning notification.
        :return:  Enables/Disables snapshot space usage threshold warning for a given volume.
        """

        return self.client.call('Network_Storage', 'setSnapshotNotification', enable, id=volume_id)

    def get_volume_snapshot_notification_status(self, volume_id):
        """returns Enabled/Disabled status of snapshot space usage threshold warning for a given volume.

        :param volume_id: ID of volume.
        :return:  Enables/Disables snapshot space usage threshold warning for a given volume.
        """
        status = self.client.call('Network_Storage', 'getSnapshotNotificationStatus', id=volume_id)
        # A None status is enabled as well.
        if status is None:
            status = 1
        # We need to force int on the return because otherwise the API will return the string '0'
        # instead of either a boolean or real int...
        return int(status)

    def authorize_host_to_volume(self, volume_id, hardware_ids=None, virtual_guest_ids=None, ip_address_ids=None,
                                 subnet_ids=None):
        """Authorizes hosts to Storage Volumes

        :param volume_id: The File volume to authorize hosts to
        :param hardware_ids: A List of SoftLayer_Hardware ids
        :param virtual_guest_ids: A List of SoftLayer_Virtual_Guest ids
        :param ip_address_ids: A List of SoftLayer_Network_Subnet_IpAddress ids
        :param subnet_ids: A List of SoftLayer_Network_Subnet ids. Only use with File volumes.
        :return: Returns an array of SoftLayer_Network_Storage_Allowed_Host objects
                which now have access to the given volume
        """
        host_templates = storage_utils.populate_host_templates(
            hardware_ids, virtual_guest_ids, ip_address_ids, subnet_ids)

        return self.client.call('Network_Storage', 'allowAccessFromHostList', host_templates, id=volume_id)

    def deauthorize_host_to_volume(self,
                                   volume_id,
                                   hardware_ids=None,
                                   virtual_guest_ids=None,
                                   ip_address_ids=None,
                                   subnet_ids=None):
        """Revokes authorization of hosts to File Storage Volumes

        :param volume_id: The File volume to deauthorize hosts to
        :param hardware_ids: A List of SoftLayer_Hardware ids
        :param virtual_guest_ids: A List of SoftLayer_Virtual_Guest ids
        :param ip_address_ids: A List of SoftLayer_Network_Subnet_IpAddress ids
        :param subnet_ids: A List of SoftLayer_Network_Subnet ids. Only use with File volumes
        :return: Returns an array of SoftLayer_Network_Storage_Allowed_Host objects
                which have access to the given File volume
        """
        host_templates = storage_utils.populate_host_templates(
            hardware_ids, virtual_guest_ids, ip_address_ids, subnet_ids)

        return self.client.call('Network_Storage', 'removeAccessFromHostList', host_templates, id=volume_id)

    def get_replication_partners(self, volume_id):
        """Acquires list of replicant volumes pertaining to the given volume.

        :param volume_id: The ID of the primary volume to be replicated
        :return: Returns an array of SoftLayer_Location objects
        """
        return self.client.call('Network_Storage', 'getReplicationPartners', id=volume_id)

    def get_replication_locations(self, volume_id):
        """Acquires list of the datacenters to which a volume can be replicated.

        :param volume_id: The ID of the primary volume to be replicated
        :return: Returns an array of SoftLayer_Network_Storage objects
        """
        return self.client.call('Network_Storage', 'getValidReplicationTargetDatacenterLocations', id=volume_id)

    def order_replicant_volume(self,
                               volume_id,
                               snapshot_schedule,
                               location,
                               tier=None,
                               os_type=None,
                               iops=None):
        """Places an order for a replicant volume.

        :param volume_id: The ID of the primary volume to be replicated
        :param snapshot_schedule: The primary volume's snapshot
                                  schedule to use for replication
        :param location: The location for the ordered replicant volume
        :param tier: The tier (IOPS per GB) of the primary volume
        :param os_type: The OS type of the primary volume
        :return: Returns a SoftLayer_Container_Product_Order_Receipt
        """

        block_mask = 'billingItem[activeChildren,hourlyFlag],' \
                     'storageTierLevel,osType,staasVersion,' \
                     'hasEncryptionAtRest,snapshotCapacityGb,schedules,' \
                     'intervalSchedule,hourlySchedule,dailySchedule,' \
                     'weeklySchedule,storageType[keyName],provisionedIops'
        block_volume = self.get_volume_details(volume_id, mask=block_mask)

        storage_class = storage_utils.block_or_file(
            block_volume['storageType']['keyName'])

        if iops is None:
            iops = int(block_volume['provisionedIops'])

        order = storage_utils.prepare_replicant_order_object(
            self, snapshot_schedule, location, tier, block_volume,
            storage_class, iops)

        if storage_class == 'block':
            if os_type is None:
                if isinstance(utils.lookup(block_volume, 'osType', 'keyName'), str):
                    os_type = block_volume['osType']['keyName']
                else:
                    raise exceptions.SoftLayerError(
                        "Cannot find primary volume's os-type automatically; must specify manually")
            order['osFormatType'] = {'keyName': os_type}

        return self.client.call('Product_Order', 'placeOrder', order)

    def order_duplicate_volume(self,
                               origin_volume_id,
                               origin_snapshot_id=None,
                               duplicate_size=None,
                               duplicate_iops=None,
                               duplicate_tier_level=None,
                               duplicate_snapshot_size=None,
                               hourly_billing_flag=False,
                               dependent_duplicate=False):
        """Places an order for a duplicate volume.

        :param origin_volume_id: The ID of the origin volume to be duplicated
        :param origin_snapshot_id: Origin snapshot ID to use for duplication
        :param duplicate_size: Size/capacity for the duplicate volume
        :param duplicate_iops: The IOPS per GB for the duplicate volume
        :param duplicate_tier_level: Tier level for the duplicate volume
        :param duplicate_snapshot_size: Snapshot space size for the duplicate
        :param hourly_billing_flag: Billing type, monthly (False) or hourly (True), default to monthly.
        :return: Returns a SoftLayer_Container_Product_Order_Receipt
        """

        block_mask = 'id,billingItem[location,hourlyFlag],snapshotCapacityGb,' \
                     'storageType[keyName],capacityGb,originalVolumeSize,' \
                     'provisionedIops,storageTierLevel,osType[keyName],' \
                     'staasVersion,hasEncryptionAtRest'
        origin_volume = self.get_volume_details(origin_volume_id,  mask=block_mask)
        storage_class = storage_utils.block_or_file(origin_volume['storageType']['keyName'])

        order = storage_utils.prepare_duplicate_order_object(
            self, origin_volume, duplicate_iops, duplicate_tier_level,
            duplicate_size, duplicate_snapshot_size, storage_class,
            hourly_billing_flag)

        if storage_class == 'block':
            if isinstance(utils.lookup(origin_volume, 'osType', 'keyName'), str):
                os_type = origin_volume['osType']['keyName']
            else:
                raise exceptions.SoftLayerError("Cannot find origin volume's os-type")

            order['osFormatType'] = {'keyName': os_type}

        if origin_snapshot_id is not None:
            order['duplicateOriginSnapshotId'] = origin_snapshot_id
        if dependent_duplicate:
            # if isDependentDuplicateFlag is set to ANYTHING, it is considered dependent.
            order['isDependentDuplicateFlag'] = 1

        return self.client.call('Product_Order', 'placeOrder', order)

    def order_modified_volume(self, volume_id, new_size=None, new_iops=None, new_tier_level=None):
        """Places an order for modifying an existing block volume.

        :param volume_id: The ID of the volume to be modified
        :param new_size: The new size/capacity for the volume
        :param new_iops: The new IOPS for the volume
        :param new_tier_level: The new tier level for the volume
        :return: Returns a SoftLayer_Container_Product_Order_Receipt
        """

        mask_items = [
            'id',
            'billingItem',
            'storageType[keyName]',
            'capacityGb',
            'provisionedIops',
            'storageTierLevel',
            'staasVersion',
            'hasEncryptionAtRest',
        ]
        block_mask = ','.join(mask_items)
        volume = self.get_volume_details(volume_id, mask=block_mask)

        order = storage_utils.prepare_modify_order_object(self, volume, new_iops, new_tier_level, new_size)

        return self.client.call('Product_Order', 'placeOrder', order)

    def volume_set_note(self, volume_id, note):
        """Set the notes for an existing block volume.

        :param volume_id: The ID of the volume to be modified
        :param note: the note
        :return: Returns true if success
        """
        template = {'notes': note}
        return self.client.call('SoftLayer_Network_Storage', 'editObject', template, id=volume_id)

    def delete_snapshot(self, snapshot_id):
        """Deletes the specified snapshot object.

        :param snapshot_id: The ID of the snapshot object to delete.
        """
        return self.client.call('Network_Storage', 'deleteObject', id=snapshot_id)

    def create_snapshot(self, volume_id, notes='', **kwargs):
        """Creates a snapshot on the given block volume.

        :param integer volume_id: The id of the volume
        :param string notes: The notes or "name" to assign the snapshot
        :return: Returns the id of the new snapshot
        """
        return self.client.call('Network_Storage', 'createSnapshot', notes, id=volume_id, **kwargs)

    def order_snapshot_space(self, volume_id, capacity, tier, upgrade,
                             iops=None, **kwargs):
        """Orders snapshot space for the given block volume.

        :param integer volume_id: The id of the volume
        :param integer capacity: The capacity to order, in GB
        :param float tier: The tier level of the block volume, in IOPS per GB
        :param boolean upgrade: Flag to indicate if this order is an upgrade
        :return: Returns a SoftLayer_Container_Product_Order_Receipt
        """
        object_mask = 'id,billingItem[location,hourlyFlag],' \
                      'storageType[keyName],storageTierLevel,provisionedIops,' \
                      'staasVersion,hasEncryptionAtRest'
        volume = self.get_volume_details(volume_id, mask=object_mask, **kwargs)

        if iops is None:
            iops = int(volume['provisionedIops'])

        order = storage_utils.prepare_snapshot_order_object(
            self, volume, capacity, tier, upgrade, iops)

        return self.client.call('Product_Order', 'placeOrder', order)

    def cancel_snapshot_space(self, volume_id, reason='No longer needed', immediate=False):
        """Cancels snapshot space for a given volume.

        :param integer volume_id: The volume ID
        :param string reason: The reason for cancellation
        :param boolean immediate_flag: Cancel immediately or on anniversary date
        """

        object_mask = 'mask[id,billingItem[activeChildren,hourlyFlag]]'
        volume = self.get_volume_details(volume_id, mask=object_mask)

        if 'activeChildren' not in volume['billingItem']:
            raise exceptions.SoftLayerError('No snapshot space found to cancel')

        children_array = volume['billingItem']['activeChildren']
        billing_item_id = None

        for child in children_array:
            if child['categoryCode'] == 'storage_snapshot_space':
                billing_item_id = child['id']
                break

        if not billing_item_id:
            raise exceptions.SoftLayerError('No snapshot space found to cancel')

        if utils.lookup(volume, 'billingItem', 'hourlyFlag'):
            immediate = True

        return self.client.call('SoftLayer_Billing_Item', 'cancelItem', immediate, True, reason, id=billing_item_id)

    def enable_snapshots(self, volume_id, schedule_type, retention_count,
                         minute, hour, day_of_week, **kwargs):
        """Enables snapshots for a specific block volume at a given schedule

        :param integer volume_id: The id of the volume
        :param string schedule_type: 'HOURLY'|'DAILY'|'WEEKLY'
        :param integer retention_count: Number of snapshots to be kept
        :param integer minute: Minute when to take snapshot
        :param integer hour: Hour when to take snapshot
        :param string day_of_week: Day when to take snapshot
        :return: Returns whether successfully scheduled or not
        """
        return self.client.call('Network_Storage', 'enableSnapshots', schedule_type, retention_count, minute,
                                hour, day_of_week, id=volume_id, **kwargs)

    def disable_snapshots(self, volume_id, schedule_type):
        """Disables snapshots for a specific block volume at a given schedule

        :param integer volume_id: The id of the volume
        :param string schedule_type: 'HOURLY'|'DAILY'|'WEEKLY'
        :return: Returns whether successfully disabled or not
        """
        return self.client.call('Network_Storage', 'disableSnapshots', schedule_type, id=volume_id)

    def list_volume_schedules(self, volume_id):
        """Lists schedules for a given volume

        :param integer volume_id: The id of the volume
        :return: Returns list of schedules assigned to a given volume
        """
        object_mask = 'schedules[type,properties[type]]'
        volume_detail = self.client.call('Network_Storage', 'getObject', id=volume_id, mask=object_mask)

        return utils.lookup(volume_detail, 'schedules')

    def restore_from_snapshot(self, volume_id, snapshot_id):
        """Restores a specific volume from a snapshot

        :param integer volume_id: The id of the volume
        :param integer snapshot_id: The id of the restore point
        :return: Returns whether succesfully restored or not
        """
        return self.client.call('Network_Storage', 'restoreFromSnapshot', snapshot_id, id=volume_id)

    def failover_to_replicant(self, volume_id, replicant_id):
        """Failover to a volume replicant.

        :param integer volume_id: The id of the volume
        :param integer replicant_id: ID of replicant to failover to
        :return: Returns whether failover was successful or not
        """
        return self.client.call('Network_Storage', 'failoverToReplicant', replicant_id, id=volume_id)

    def disaster_recovery_failover_to_replicant(self, volume_id, replicant_id):
        """Disaster Recovery Failover to a volume replicant.

        :param integer volume_id: The id of the volume
        :param integer replicant: ID of replicant to failover to
        :return: Returns whether failover to successful or not
        """
        return self.client.call('Network_Storage', 'disasterRecoveryFailoverToReplicant', replicant_id, id=volume_id)

    def failback_from_replicant(self, volume_id):
        """Failback from a volume replicant.

        :param integer volume_id: The id of the volume
        :return: Returns whether failback was successful or not
        """
        return self.client.call('Network_Storage', 'failbackFromReplicant', id=volume_id)

    def cancel_volume(self, volume_id, reason='No longer needed', immediate=False):
        """Cancels the given storage volume.

        :param integer volume_id: The volume ID
        :param string reason: The reason for cancellation
        :param boolean immediate_flag: Cancel immediately or on anniversary date
        """
        object_mask = 'mask[id,billingItem[id,hourlyFlag]]'
        volume = self.get_volume_details(volume_id, mask=object_mask)

        if 'billingItem' not in volume:
            raise exceptions.SoftLayerError("Storage Volume was already cancelled")

        billing_item_id = volume['billingItem']['id']

        if utils.lookup(volume, 'billingItem', 'hourlyFlag'):
            immediate = True

        return self.client.call('SoftLayer_Billing_Item', 'cancelItem', immediate, True, reason, id=billing_item_id)

    def refresh_dupe(self, volume_id, snapshot_id, force_refresh=False):
        """"Refresh a duplicate volume with a snapshot from its parent.

        :param integer volume_id: The id of the volume
        :param integer snapshot_id: The id of the snapshot
        :param boolean force_refresh: Force refreshing the volume if True
        """
        return self.client.call('Network_Storage', 'refreshDuplicate', snapshot_id, force_refresh, id=volume_id)

    def convert_dep_dupe(self, volume_id):
        """Convert a dependent duplicate volume to an independent volume.

        :param integer volume_id: The id of the volume.
        """
        return self.client.call('Network_Storage', 'convertCloneDependentToIndependent', id=volume_id)

    def convert_dupe_status(self, volume_id):
        """Get the Clone split/move status completion of a duplicate volume

        :param integer volume_id: The id of the volume.
        """
        return self.client.call('Network_Storage', 'getDuplicateConversionStatus', id=volume_id)

    def get_network_message_delivery_accounts(self, object_id):
        """Return  object data of the cloud storage.

        :param object_id cloud object storage identifier
        Returns: Get instances
        """
        object_mask = 'mask[uuid,credentials]'
        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'getObject',
                                mask=object_mask, id=object_id)

    def get_end_points(self, object_id):
        """Returns a collection of endpoint URLs available to this IBM Cloud Object Storage account.

        :param object_id cloud object storage identifier
        Returns: Returns a collection of endpoint URLs.
        """
        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'getEndpoints',  id=object_id)



"""
    SoftLayer.testing.xmlrpc
    ~~~~~~~~~~~~~~~~~~~~~~~~
    XMP-RPC server which can use a transport to proxy requests for testing.

    If you want to spin up a test XML server to make fake API calls with, try this:

    quick-server.py
    ---
    import SoftLayer
    from SoftLayer.testing import xmlrpc

    my_xport = SoftLayer.FixtureTransport()
    my_server = xmlrpc.create_test_server(my_xport, "localhost", port=4321)
    print(f"Server running on http://{my_server.server_name}:{my_server.server_port}")
    ---
    $> python quick-server.py
    $> curl -X POST -d "<?xml version='1.0' encoding='iso-8859-1'?><methodCall><methodName> \
getInvoiceTopLevelItems</methodName><params><param><value><struct><member><name>headers</name> \
<value><struct><member><name>SoftLayer_Billing_InvoiceInitParameters</name><value><struct> \
<member><name>id</name><value><string>1234</string></value></member></struct></value></member> \
</struct></value></member></struct></value></param></params></methodCall>" \
http://127.0.0.1:4321/SoftLayer_Billing_Invoice

    :license: MIT, see LICENSE for more details.
"""
import http.server
import logging
import threading
import xmlrpc.client

import SoftLayer
from SoftLayer import transports
from SoftLayer import utils

# pylint: disable=invalid-name, broad-except, arguments-differ


class TestServer(http.server.ThreadingHTTPServer):
    """Test HTTP server which holds a given transport."""

    def __init__(self, transport, *args, **kw):
        http.server.ThreadingHTTPServer.__init__(self, *args, **kw)
        self.transport = transport


class TestHandler(http.server.BaseHTTPRequestHandler):
    """Test XML-RPC Handler which converts XML-RPC to transport requests."""

    def do_POST(self):
        """Handle XML-RPC POSTs."""
        try:
            length = int(self.headers['Content-Length'])
            data = self.rfile.read(length).decode('utf-8')
            args, method = xmlrpc.client.loads(data)
            headers = args[0].get('headers', {})

            # Form Request for the transport
            req = transports.Request()
            req.service = self.path.lstrip('/')
            req.method = method
            req.limit = utils.lookup(headers, 'resultLimit', 'limit')
            req.offset = utils.lookup(headers, 'resultLimit', 'offset')
            req.args = args[1:]
            req.filter = _item_by_key_postfix(headers, 'ObjectFilter') or None
            req.mask = _item_by_key_postfix(headers, 'ObjectMask').get('mask')
            req.identifier = _item_by_key_postfix(headers, 'InitParameters').get('id')
            req.transport_headers = dict(((k.lower(), v) for k, v in self.headers.items()))
            req.headers = headers

            # Get response
            response = self.server.transport(req)

            # Need to convert BACK to list, so xmlrpc can dump it out properly.
            if isinstance(response, SoftLayer.transports.transport.SoftLayerListResult):
                response = list(response)
            response_body = xmlrpc.client.dumps((response,), allow_none=True, methodresponse=True)

            self.send_response(200)
            self.send_header("Content-type", "application/xml; charset=UTF-8")
            self.end_headers()

            try:
                self.wfile.write(response_body.encode('utf-8'))
            except UnicodeDecodeError:
                self.wfile.write(response_body)

        except (NotImplementedError, NameError) as ex:
            self.send_response(200)
            self.end_headers()
            response = xmlrpc.client.Fault(404, str(ex))
            response_body = xmlrpc.client.dumps(response, allow_none=True, methodresponse=True)
            self.wfile.write(response_body.encode('utf-8'))

        except SoftLayer.SoftLayerAPIError as ex:
            self.send_response(200)
            self.end_headers()
            response = xmlrpc.client.Fault(ex.faultCode, str(ex.reason))
            response_body = xmlrpc.client.dumps(response, allow_none=True, methodresponse=True)
            self.wfile.write(response_body.encode('utf-8'))
        except OverflowError as ex:
            self.send_response(555)
            self.send_header("Content-type", "application/xml; charset=UTF-8")
            self.end_headers()
            response_body = '''<error>OverflowError in XML response.</error>'''
            self.wfile.write(response_body.encode('utf-8'))
            logging.exception("Error while handling request: %s", ex)
        except Exception as ex:
            self.send_response(500)
            logging.exception("Error while handling request: %s", ex)

    def log_message(self, fmt, *args):
        """Override log_message."""


def _item_by_key_postfix(dictionary, key_prefix):
    """Get item from a dictionary which begins with the given prefix."""
    for key, value in dictionary.items():
        if key.endswith(key_prefix):
            return value

    return {}


def create_test_server(transport, host='localhost', port=0):
    """Create a test XML-RPC server in a new thread."""
    server = TestServer(transport, (host, port), TestHandler)
    thread = threading.Thread(target=server.serve_forever, kwargs={'poll_interval': 0.01})
    thread.start()
    return server



"""Display permission details for a cloud object storage."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('object_id')
@environment.pass_env
def cli(env, object_id):
    """Display permission details for a cloud object storage."""

    block_manager = SoftLayer.BlockStorageManager(env.client)

    cloud = block_manager.get_network_message_delivery_accounts(object_id)
    end_points = block_manager.get_end_points(object_id)

    table = formatting.Table(['Name', 'Value'])

    table_credentials = formatting.Table(['Id', 'Access Key ID', 'Secret Access Key', 'Description'])

    for credential in cloud.get('credentials'):
        table_credentials.add_row([credential.get('id'),
                                   credential.get('username'),
                                   credential.get('password'),
                                   credential['type']['description']])

    table_url = formatting.Table(['Region',
                                  'Location',
                                  'Type',
                                  'URL'])
    for end_point in end_points:
        table_url.add_row([end_point.get('region') or '',
                           end_point.get('location') or '',
                           end_point.get('type'),
                           end_point.get('url'), ])

    table.add_row(['UUID', cloud.get('uuid')])
    table.add_row(['Credentials', table_credentials])
    table.add_row(['EndPoint URL´s', table_url])
    env.fout(table)



"""Update firmware."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('-i', '--ipmi', is_flag=True, help="Update IPMI firmware")
@click.option('-r', '--raid', is_flag=True, help="Update RAID firmware")
@click.option('-b', '--bios', is_flag=True, help="Update BIOS firmware")
@click.option('-d', '--harddrive', is_flag=True, help="Update Hard Drives firmware")
@click.option('-n', '--network', is_flag=True, help="Update Network Card firmware")
@environment.pass_env
def cli(env, identifier, ipmi, raid, bios, harddrive, network):
    """Update server firmware. By default will update all available server components."""

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')
    confirm_message = f"This will power off the server with id {hw_id} and update device firmware. Continue?"
    if not (env.skip_confirmations or formatting.confirm(confirm_message)):
        raise exceptions.CLIAbort('Aborted.')

    # If no options were specified, set them all to enabled.
    if not any([ipmi, raid, bios, harddrive, network]):
        ipmi = raid = bios = harddrive = network = 1
    mgr.update_firmware(hw_id, ipmi, raid, bios, harddrive, network)
    env.fout(f"[green]Firmware update for {identifier} started")



"""Get Load Balancer as a Service details."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get Load Balancer as a Service details."""
    mgr = SoftLayer.LoadBalancerManager(env.client)
    _, lbid = mgr.get_lbaas_uuid_id(identifier)
    this_lb = mgr.get_lb(lbid)
    if this_lb.get('previousErrorText'):
        print(this_lb.get('previousErrorText'))
    table = lbaas_table(this_lb)

    env.fout(table)


def lbaas_table(this_lb):
    """Generates a table from a list of LBaaS devices"""
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Id', this_lb.get('id')])
    table.add_row(['UUI', this_lb.get('uuid')])
    table.add_row(['Name', this_lb.get('name')])
    table.add_row(['Address', this_lb.get('address')])
    table.add_row(['Type', SoftLayer.LoadBalancerManager.TYPE.get(this_lb.get('type'))])
    table.add_row(['Location', utils.lookup(this_lb, 'datacenter', 'longName')])
    table.add_row(['Description', this_lb.get('description')])
    table.add_row(['Status', f"{this_lb.get('provisioningStatus')} / {this_lb.get('operatingStatus')}"])

    listener_table, pools = get_listener_table(this_lb)
    table.add_row(['Protocols', listener_table])

    if pools.get('members') is not None:
        member_table = get_member_table(this_lb, pools)
        table.add_row(['Members', member_table])
    else:
        table.add_row(['Members', "Not Found"])

    if this_lb.get('healthMonitors') != []:
        hp_table = get_hp_table(this_lb)
        table.add_row(['Health Checks', hp_table])
    else:
        table.add_row(['Health Checks', "Not Found"])

    if this_lb.get('l7Pools') != []:
        l7pool_table = get_l7pool_table(this_lb)
        table.add_row(['L7 Pools', l7pool_table])
    else:
        table.add_row(['L7 Pools', "Not Found"])

    if this_lb.get('sslCiphers') != []:
        ssl_table = get_ssl_table(this_lb)
        table.add_row(['Ciphers', ssl_table])
    else:
        table.add_row(['Ciphers', "Not Found"])

    return table


def get_hp_table(this_lb):
    """Generates a table from a list of LBaaS devices"""
    # https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_HealthMonitor/
    hp_table = formatting.Table(['UUID', 'Interval', 'Retries', 'Type', 'ServerTimeout ', 'Modify', 'Active'])
    for health in this_lb.get('healthMonitors', []):
        hp_table.add_row([
            health.get('uuid'),
            health.get('interval'),
            health.get('maxRetries'),
            health.get('monitorType'),
            health.get('serverTimeout'),
            utils.clean_time(health.get('modifyDate')),
            health.get('provisioningStatus')
        ])
    return hp_table


def get_member_table(this_lb, pools):
    """Generates a members table from a list of LBaaS devices"""
    # https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_Member/
    member_col = ['UUID', 'Address', 'Weight', 'Modify', 'Active']
    counter = 0
    for uuid in pools.values():
        member_col.append(f'P{counter}-> {uuid}')
        counter += 1
    member_table = formatting.Table(member_col)
    for member in this_lb.get('members', []):
        row = [
            member.get('uuid'),
            member.get('address'),
            member.get('weight'),
            utils.clean_time(member.get('modifyDate')),
            member.get('provisioningStatus')
        ]
        for uuid in pools:
            row.append(get_member_hp(this_lb.get('health'), member.get('uuid'), uuid))
        member_table.add_row(row)
    return member_table


def get_ssl_table(this_lb):
    """Generates a ssl table from a list of LBaaS devices"""
    # https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_SSLCipher/
    ssl_table = formatting.Table(['Id', 'Name'])
    for ssl in this_lb.get('sslCiphers', []):
        ssl_table.add_row([ssl.get('id'), ssl.get('name')])
    return ssl_table


def get_listener_table(this_lb):
    """Generates a protocols table from a list of LBaaS devices"""
    pools = {}
    # https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_Listener/
    listener_table = formatting.Table(['UUID', 'Max Connection', 'Mapping', 'Balancer', 'Modify', 'Active'])
    for listener in this_lb.get('listeners', []):
        pool = listener.get('defaultPool')
        priv_map = f"{pool['protocol']}:{pool['protocolPort']}"
        pools[pool['uuid']] = priv_map
        mapping = f"{listener.get('protocol')}:{listener.get('protocolPort')} -> {priv_map}"
        listener_table.add_row([
            listener.get('uuid'),
            listener.get('connectionLimit', 'None'),
            mapping,
            pool.get('loadBalancingAlgorithm', 'None'),
            utils.clean_time(listener.get('modifyDate')),
            listener.get('provisioningStatus')
        ])
    return listener_table, pools


def get_l7pool_table(this_lb):
    """Generates a l7Pools table from a list of LBaaS devices"""
    # https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_L7Pool/
    l7_table = formatting.Table(['Id', 'UUID', 'Balancer', 'Name', 'Protocol', 'Modify', 'Active'])
    for layer7 in this_lb.get('l7Pools', []):
        l7_table.add_row([
            layer7.get('id'),
            layer7.get('uuid'),
            layer7.get('loadBalancingAlgorithm'),
            layer7.get('name'),
            layer7.get('protocol'),
            utils.clean_time(layer7.get('modifyDate')),
            layer7.get('provisioningStatus')
        ])
    return l7_table


def get_member_hp(checks, member_uuid, pool_uuid):
    """Helper function to find a members health in a given pool

    :param checks list: https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_Pool/#healthMonitor
    :param member_uuid: server UUID we are looking for
    :param pool_uuid: Connection pool uuid to search for
    """
    status = "---"
    for check in checks:
        if check.get('poolUuid') == pool_uuid:
            for hp_member in check.get('membersHealth'):
                if hp_member.get('uuid') == member_uuid:
                    status = hp_member.get('status')

    return status



"""Invoice details"""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.argument('identifier')
@click.option('--details', is_flag=True, default=False, show_default=True,
              help="Shows a very detailed list of charges")
@environment.pass_env
def cli(env, identifier, details):
    """Invoice details

    Will display the top level invoice items for a given invoice. The cost displayed is the sum of the item's
    cost along with all its child items.
    The --details option will display any child items a top level item may have. Parent items will appear
    in this list as well to display their specific cost.
    """

    manager = AccountManager(env.client)
    top_items = manager.get_billing_items(identifier)
    table = get_invoice_table(identifier, top_items, details)
    env.fout(table)


def nice_string(ugly_string, limit=100):
    """Format and trims strings"""
    return (ugly_string[:limit] + '..') if len(ugly_string) > limit else ugly_string


def get_invoice_table(identifier, top_items, details):
    """Formats a table for invoice top level items.

     :param int identifier: Invoice identifier.
     :param list top_items: invoiceTopLevelItems.
     :param bool details: To add very detailed list of charges.
      """

    title = "Invoice %s" % identifier
    table = formatting.Table(["Item Id", "Category", "Description", "Single",
                              "Monthly", "Create Date", "Location"], title=title)
    table.align['category'] = 'l'
    table.align['description'] = 'l'
    for item in top_items:
        fqdn = "%s.%s" % (item.get('hostName', ''), item.get('domainName', ''))
        # category id=2046, ram_usage doesn't have a name...
        category = utils.lookup(item, 'category', 'name') or item.get('categoryCode')
        description = nice_string(item.get('description'))
        if fqdn != '.':
            description = "%s (%s)" % (item.get('description'), fqdn)
        total_recur, total_single = sum_item_charges(item)
        table.add_row([
            item.get('id'),
            category,
            nice_string(description),
            f"${total_single:,.2f}",
            f"${total_recur:,.2f}",
            utils.clean_time(item.get('createDate'), out_format="%Y-%m-%d"),
            utils.lookup(item, 'location', 'name')
        ])
        if details:
            # This item has children, so we want to print out the parent item too. This will match the
            # invoice from the portal. https://github.com/softlayer/softlayer-python/issues/2201
            if len(item.get('children')) > 0:
                single = float(item.get('oneTimeAfterTaxAmount', 0.0))
                recurring = float(item.get('recurringAfterTaxAmount', 0.0))
                table.add_row([
                    '>>>',
                    category,
                    nice_string(description),
                    f"${single:,.2f}",
                    f"${recurring:,.2f}",
                    '---',
                    '---'
                ])
            for child in item.get('children', []):
                table.add_row([
                    '>>>',
                    utils.lookup(child, 'category', 'name'),
                    nice_string(child.get('description')),
                    "$%.2f" % float(child.get('oneTimeAfterTaxAmount')),
                    "$%.2f" % float(child.get('recurringAfterTaxAmount')),
                    '---',
                    '---'
                ])
    return table


def sum_item_charges(item: dict) -> (float, float):
    """Takes a billing Item, sums up its child items and returns recurring, one_time prices"""

    # API returns floats as strings in this case
    single = float(item.get('oneTimeAfterTaxAmount', 0.0))
    recurring = float(item.get('recurringAfterTaxAmount', 0.0))
    for child in item.get('children', []):
        single = single + float(child.get('oneTimeAfterTaxAmount', 0.0))
        recurring = recurring + float(child.get('recurringAfterTaxAmount', 0.0))

    return (recurring, single)



"""
    SoftLayer.CLI.environment
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    Abstracts everything related to the user's environment when running the CLI

    :license: MIT, see LICENSE for more details.
"""
import configparser
import os

import importlib
from json.decoder import JSONDecodeError

import click

from rich.console import Console
from rich.syntax import Syntax

import SoftLayer
from SoftLayer.CLI import formatting
from SoftLayer.CLI import routes
from SoftLayer import utils

# pylint: disable=too-many-instance-attributes, invalid-name


class Environment(object):
    """Provides access to the current CLI environment."""

    def __init__(self):
        # {'path:to:command': ModuleLoader()}
        # {'vs:list': ModuleLoader()}
        self.commands = {}
        self.aliases = {}

        self.vars = {}

        self.client = None
        self.theme = self.set_env_theme()
        self.console = utils.console_color_themes(self.theme)
        self.err_console = Console(stderr=True)
        self.format = 'table'
        self.skip_confirmations = False
        self.config_file = None

        self._modules_loaded = False

    def out(self, output):
        """Outputs a string to the console (stdout)."""

        # If we output to a | or file, need to set default width so all output is printed.
        if not self.console.is_terminal:
            self.console.width = 1000000
        if self.format == 'json':
            try:
                self.console.print_json(output)
            # Tried to print not-json, so just print it out normally...
            except JSONDecodeError:
                click.echo(output)
        elif self.format == 'jsonraw':
            #  Using Rich here is problematic because in the unit tests it thinks the terminal is 80 characters wide
            #  and only prints out that many characters.
            click.echo(output)
        else:
            # If we want to print a list of tables, Rich doens't handle that well.
            if isinstance(output, list):
                for line in output:
                    self.console.print(line, overflow='ignore')
            else:
                self.console.print(output,  overflow='ignore')

    def err(self, output, newline=True):
        """Outputs an error string to the console (stderr)."""

        self.err_console.print(output, new_line_start=newline)

    def fmt(self, output, fmt=None):
        """Format output based on current the environment format."""
        if fmt is None:
            fmt = self.format
        return formatting.format_output(output, fmt, self.theme)

    def format_output_is_json(self):
        """Return True if format output is json or jsonraw"""
        return 'json' in self.format

    def fout(self, output):
        """Format the input and output to the console (stdout)."""
        if output is not None:
            try:
                self.out(self.fmt(output))
            except UnicodeEncodeError:
                # If we hit an undecodeable entry, just try outputting as json.
                self.out(self.fmt(output, 'json'))

    def python_output(self, output):
        """Prints out python code"""
        self.console.print(Syntax(output, "python"))

    def input(self, prompt, default=None, show_default=True):
        """Provide a command prompt."""
        return click.prompt(prompt, default=default, show_default=show_default)

    def getpass(self, prompt, default=None):
        """Provide a password prompt."""
        password = click.prompt(prompt, hide_input=True, default=default)

        # https://github.com/softlayer/softlayer-python/issues/1436
        # click.prompt uses python's getpass() in the background
        # https://github.com/python/cpython/blob/3.9/Lib/getpass.py#L97
        # In windows, shift+insert actually inputs the below 2 characters
        # If we detect those 2 characters, need to manually read from the clipbaord instead
        # https://stackoverflow.com/questions/101128/how-do-i-read-text-from-the-clipboard
        # LINUX NOTICE: `apt-get install python3-tk` required to install tk
        if password == 'àR':
            # tkinter is a built in python gui, but it has clipboard reading functions.
            # pylint: disable=import-outside-toplevel
            from tkinter import Tk
            tk_manager = Tk()
            password = tk_manager.clipboard_get()
            # keep the window from showing
            tk_manager.withdraw()
        return password

    # Command loading methods
    def list_commands(self, *path):
        """Command listing."""
        path_str = ':'.join(path)

        commands = []
        for command in self.commands:

            # Filter based on prefix and the segment length
            if all([command.startswith(path_str),
                    len(path) == command.count(":")]):

                # offset is used to exclude the path that the caller requested.
                offset = len(path_str) + 1 if path_str else 0
                if ':' not in command[offset:]:
                    commands.append(command[offset:])

        return sorted(commands)

    def get_command(self, *path):
        """Return command at the given path or raise error."""
        path_str = ':'.join(path)

        if path_str in self.commands:
            return self.commands[path_str].load()

        return None

    def resolve_alias(self, path_str):
        """Returns the actual command name. Uses the alias mapping."""
        if path_str in self.aliases:
            return self.aliases[path_str]
        return path_str

    def load(self):
        """Loads all modules."""
        if self._modules_loaded is True:
            return

        self.load_modules_from_python(routes.ALL_ROUTES)
        self.aliases.update(routes.ALL_ALIASES)
        self._modules_loaded = True

    def load_modules_from_python(self, route_list):
        """Load modules from the native python source."""
        for name, modpath in route_list:
            if ':' in modpath:
                path, attr = modpath.split(':', 1)
            else:
                path, attr = modpath, None
            self.commands[name] = ModuleLoader(path, attr=attr)

    def ensure_client(self, config_file=None, is_demo=False, proxy=None):
        """Create a new SLAPI client to the environment.

        This will be a no-op if there is already a client in this environment.
        """
        if self.client is not None:
            return

        # Environment can be passed in explicitly. This is used for testing
        if is_demo:
            client = SoftLayer.BaseClient(transport=SoftLayer.FixtureTransport(), auth=None)
        else:
            # Create SL Client
            client = SoftLayer.create_client_from_env(proxy=proxy, config_file=config_file)
        self.client = client

    def ensure_emp_client(self, config_file=None, is_demo=False, proxy=None):
        """Create a new SLAPI client to the environment.

        This will be a no-op if there is already a client in this environment.
        """
        if self.client is not None:
            return

        # Environment can be passed in explicitly. This is used for testing
        if is_demo:
            client = SoftLayer.BaseClient(transport=SoftLayer.FixtureTransport(), auth=None)
        else:
            # Create SL Client
            client = SoftLayer.employee_client(proxy=proxy, config_file=config_file)
        self.client = client

    def set_env_theme(self, config_file=None):
        """Get theme to color console and set in env"""
        theme = os.environ.get('SL_THEME')
        if theme:
            return theme
        else:
            config_files = ['/etc/softlayer.conf', '~/.softlayer']
            path_os = os.getenv('HOME')
            if path_os:
                config_files.append(path_os + '\\AppData\\Roaming\\softlayer')
            if config_file:
                config_files.append(config_file)
            config = configparser.RawConfigParser({'theme': 'dark'})
            config.read(config_files)
            if config.has_section('softlayer'):
                self.theme = config.get('softlayer', 'theme')
                return config.get('softlayer', 'theme')
        return 'dark'


class ModuleLoader(object):
    """Module loader that acts a little like an EntryPoint object."""

    def __init__(self, import_path, attr=None):
        self.import_path = import_path
        self.attr = attr

    def load(self):
        """load and return the module/attribute."""
        module = importlib.import_module(self.import_path)
        if self.attr:
            return getattr(module, self.attr)
        return module


pass_env = click.make_pass_decorator(Environment, ensure=True)



"""Call arbitrary API endpoints."""
import json

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils

SPLIT_TOKENS = [
    ('in', ' IN '),
    ('eq', '='),
]


def _build_filters(_filters):
    """Builds filters using the filter options passed into the CLI.

    This only supports the equals keyword at the moment.
    """
    root = utils.NestedDict({})
    for _filter in _filters:
        operation = None
        for operation, token in SPLIT_TOKENS:
            # split "some.key=value" into ["some.key", "value"]
            top_parts = _filter.split(token, 1)
            if len(top_parts) == 2:
                break
        else:
            raise exceptions.CLIAbort('Failed to find valid operation for: %s' % _filter)

        key, value = top_parts
        current = root
        # split "some.key" into ["some", "key"]
        parts = [part.strip() for part in key.split('.')]

        # Actually drill down and add the filter
        for part in parts[:-1]:
            current = current[part]

        if operation == 'eq':
            current[parts[-1]] = utils.query_filter(value.strip())
        elif operation == 'in':
            current[parts[-1]] = {
                'operation': 'in',
                'options': [{
                    'name': 'data',
                    'value': [p.strip() for p in value.split(',')],
                }],
            }

    return root.to_dict()


def _build_python_example(args, kwargs):
    sorted_kwargs = sorted(kwargs.items())

    call_str = 'import SoftLayer\n\n'
    call_str += 'client = SoftLayer.create_client_from_env()\n'
    call_str += 'result = client.call('
    arg_list = [repr(arg) for arg in args]
    arg_list += [key + '=' + repr(value)
                 for key, value in sorted_kwargs if value]
    call_str += ',\n                     '.join(arg_list)
    call_str += ')'

    return call_str


def _validate_filter(ctx, param, value):  # pylint: disable=unused-argument
    """Validates a JSON style object filter"""
    _filter = None
    if value:
        try:
            _filter = json.loads(value)
            if not isinstance(_filter, dict):
                raise exceptions.CLIAbort(f"\"{_filter}\" should be a JSON object, but is a {type(_filter)} instead.")
        except json.JSONDecodeError as error:
            raise exceptions.CLIAbort(f"\"{value}\" is not valid JSON. {error}")

    return _filter


def _validate_parameters(ctx, param, value):  # pylint: disable=unused-argument
    """Checks if value is a JSON string, and converts it to a datastructure if that is true"""

    validated_values = []
    for parameter in value:
        if isinstance(parameter, str):
            # looks like a JSON string...
            if '{' in parameter or '[' in parameter:
                try:
                    parameter = json.loads(parameter)
                except json.JSONDecodeError as error:
                    click.secho(f"{parameter} looked like json, but was invalid, passing to API as is. {error}",
                                fg='red')
        validated_values.append(parameter)
    return validated_values


@click.command('call', short_help="Call arbitrary API endpoints.", cls=SLCommand)
@click.argument('service')
@click.argument('method')
@click.argument('parameters', nargs=-1, callback=_validate_parameters)
@click.option('--id', '_id', help="Init parameter")
@helpers.multi_option('--filter', '-f', '_filters',
                      help="Object filters. This should be of the form: 'property=value' or 'nested.property=value'."
                           "Complex filters should use --json-filter.")
@click.option('--mask', help="String-based object mask")
@click.option('--limit', type=click.INT, help="Result limit")
@click.option('--offset', type=click.INT, help="Result offset")
@click.option('--orderBy', type=click.STRING,
              help="To set the sort direction, ASC or DESC can be provided."
                   "This should be of the form: '--orderBy nested.property' default DESC or "
                   "'--orderBy nested.property=ASC', e.g. "
                   " --orderBy subnets.id default DESC"
                   " --orderBy subnets.id=ASC")
@click.option('--output-python / --no-output-python',
              help="Show python example code instead of executing the call")
@click.option('--json-filter', callback=_validate_filter,
              help="A JSON string to be passed in as the object filter to the API call. "
                   "Remember to use double quotes (\") for variable names. Can NOT be used with --filter. "
                   "Dont use whitespace outside of strings, or the slcli might have trouble parsing it.")
@environment.pass_env
def cli(env, service, method, parameters, _id, _filters, mask, limit, offset, orderby=None,
        output_python=False, json_filter=None):
    """Call arbitrary API endpoints with the given SERVICE and METHOD.

    For parameters that require a datatype, use a JSON string for that parameter.
    Example::

        slcli call-api Account getObject
        slcli call-api Account getVirtualGuests --limit=10 --mask=id,hostname
        slcli call-api Virtual_Guest getObject --id=12345
        slcli call-api Metric_Tracking_Object getBandwidthData --id=1234 \\
            "2015-01-01 00:00:00" "2015-01-1 12:00:00" public
        slcli call-api Account getVirtualGuests \\
            -f 'virtualGuests.datacenter.name=dal05' \\
            -f 'virtualGuests.maxCpu=4' \\
            --mask=id,hostname,datacenter.name,maxCpu
        slcli call-api Account getVirtualGuests \\
            -f 'virtualGuests.datacenter.name IN dal05,sng01'
        slcli call-api Account getVirtualGuests \\
            --json-filter  '{"virtualGuests":{"hostname":{"operation":"^= test"}}}' --limit=10
        slcli -v call-api SoftLayer_User_Customer addBulkPortalPermission --id=1234567 \\
            '[{"keyName": "NETWORK_MESSAGE_DELIVERY_MANAGE"}]'
        slcli call-api Account getVirtualGuests \\
            --orderBy virttualguests.id=ASC
        slcli call-api SoftLayer_Notification_Occurrence_Event getAllObjects \\
            --json-filter='{"endDate": {"operation": "greaterThanDate", \\
            "options": [{"name":"date", "value": ["10/14/2022"]}]}}' --limit=50
    """

    if _filters and json_filter:
        raise exceptions.CLIAbort("--filter and --json-filter cannot be used together.")

    object_filter = _build_filters(_filters)
    if orderby:
        orderby = utils.build_filter_orderby(orderby)
        object_filter = utils.dict_merge(object_filter, orderby)
    if json_filter:
        object_filter = utils.dict_merge(json_filter, object_filter)

    args = [service, method] + list(parameters)
    kwargs = {
        'id': _id,
        'filter': object_filter,
        'mask': mask,
        'limit': limit,
        'offset': offset,
    }

    if output_python:
        env.python_output(_build_python_example(args, kwargs))
    else:
        result = env.client.call(*args, **kwargs)
        env.fout(formatting.iter_to_table(result))



"""Manage LBaaS Pools/Listeners."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers
from SoftLayer.exceptions import SoftLayerAPIError


# pylint: disable=unused-argument
def sticky_option(ctx, param, value):
    """parses sticky cli option"""
    if value:
        return 'SOURCE_IP'
    return None


def parse_server(ctx, param, values):
    """Splits out the IP, Port, Weight from the --server argument for l7pools"""
    servers = []
    for server in values:
        splitout = server.split(':')
        if len(splitout) != 3:
            raise exceptions.ArgumentError(
                f"--server needs a port and a weight. {server} improperly formatted")
        server = {
            'address': splitout[0],
            'port': splitout[1],
            'weight': splitout[2]
        }
        servers.append(server)

    return servers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--frontProtocol', '-P', default='HTTP', type=click.Choice(['HTTP', 'HTTPS', 'TCP']), show_default=True,
              help="Protocol type to use for incoming connections")
@click.option('--backProtocol', '-p', type=click.Choice(['HTTP', 'HTTPS', 'TCP']),
              help="Protocol type to use when connecting to backend servers. Defaults to whatever --frontProtocol is.")
@click.option('--frontPort', '-f', required=True, type=int, help="Internet side port")
@click.option('--backPort', '-b', required=True, type=int, help="Private side port")
@click.option('--method', '-m', default='ROUNDROBIN', show_default=True, help="Balancing Method",
              type=click.Choice(['ROUNDROBIN', 'LEASTCONNECTION', 'WEIGHTED_RR']))
@click.option('--connections', '-c', type=int, help="Maximum number of connections to allow.")
@click.option('--sticky', '-s', is_flag=True, callback=sticky_option, help="Make sessions sticky based on source_ip.")
@click.option('--sslCert', '-x', help="SSL certificate ID. See `slcli ssl list`")
@environment.pass_env
def add(env, identifier, **args):
    """Adds a listener to the identifier LB"""

    mgr = SoftLayer.LoadBalancerManager(env.client)
    uuid, _ = mgr.get_lbaas_uuid_id(identifier)

    new_listener = {
        'backendPort': args.get('backport'),
        'backendProtocol': args.get('backprotocol') if args.get('backprotocol') else args.get('frontprotocol'),
        'frontendPort': args.get('frontport'),
        'frontendProtocol': args.get('frontprotocol'),
        'loadBalancingMethod': args.get('method'),
        'maxConn': args.get('connections', None),
        'sessionType': args.get('sticky'),
        'tlsCertificateId': args.get('sslcert')
    }

    try:
        mgr.add_lb_listener(uuid, new_listener)
        click.secho("Success", fg='green')
    except SoftLayerAPIError as exception:
        click.secho(f"ERROR: {exception.faultString}", fg='red')


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.argument('listener')
@click.option('--frontProtocol', '-P', type=click.Choice(['HTTP', 'HTTPS', 'TCP']),
              help="Protocol type to use for incoming connections")
@click.option('--backProtocol', '-p', type=click.Choice(['HTTP', 'HTTPS', 'TCP']),
              help="Protocol type to use when connecting to backend servers. Defaults to whatever --frontProtocol is.")
@click.option('--frontPort', '-f', type=int, help="Internet side port")
@click.option('--backPort', '-b', type=int, help="Private side port")
@click.option('--method', '-m', help="Balancing Method",
              type=click.Choice(['ROUNDROBIN', 'LEASTCONNECTION', 'WEIGHTED_RR']))
@click.option('--connections', '-c', type=int, help="Maximum number of connections to allow.")
@click.option('--clientTimeout', '-t', type=int,
              help="maximum idle time in seconds(Range: 1 to 7200).")
@click.option('--sticky', '-s', is_flag=True, callback=sticky_option, help="Make sessions sticky based on source_ip.")
@click.option('--sslCert', '-x', help="SSL certificate ID. See `slcli ssl list`")
@environment.pass_env
def edit(env, identifier, listener, **args):
    """Updates a listener's configuration.

    LISTENER should be a UUID, and can be found from `slcli lb detail <IDENTIFIER>`
    """

    mgr = SoftLayer.LoadBalancerManager(env.client)
    uuid, _ = mgr.get_lbaas_uuid_id(identifier)

    new_listener = {
        'listenerUuid': listener
    }

    arg_to_option = {
        'frontprotocol': 'frontendProtocol',
        'backprotocol': 'backendProtocol',
        'frontport': 'frontendPort',
        'backport': 'backendPort',
        'method': 'loadBalancingMethod',
        'connections': 'maxConn',
        'sticky': 'sessionType',
        'clienttimeout': 'clientTimeout',
        'sslcert': 'tlsCertificateId'
    }

    for key, value in args.items():
        if value:
            new_listener[arg_to_option[key]] = value

    try:
        mgr.add_lb_listener(uuid, new_listener)
        click.secho("Success", fg='green')
    except SoftLayerAPIError as exception:
        click.secho(f"ERROR: {exception.faultString}", fg='red')


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.argument('listener')
@environment.pass_env
def delete(env, identifier, listener):
    """Removes the listener from identified LBaaS instance

    LISTENER should be a UUID, and can be found from `slcli lb detail <IDENTIFIER>`
    """

    mgr = SoftLayer.LoadBalancerManager(env.client)
    uuid, _ = mgr.get_lbaas_uuid_id(identifier)
    try:
        mgr.remove_lb_listener(uuid, listener)
        click.secho("Success", fg='green')
    except SoftLayerAPIError as exception:
        click.secho(f"ERROR: {exception.faultString}", fg='red')


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
# https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_L7Pool/
@click.option('--name', '-n', required=True, help="Name for this L7 pool.")
@click.option('--method', '-m', help="Balancing Method.", default='ROUNDROBIN', show_default=True,
              type=click.Choice(['ROUNDROBIN', 'LEASTCONNECTION', 'WEIGHTED_RR']))
@click.option('--protocol', '-P', type=click.Choice(['HTTP', 'HTTPS']), default='HTTP',
              show_default=True, help="Protocol type to use for incoming connections")
# https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_L7Member/
@helpers.multi_option('--server', '-S', callback=parse_server, required=True,
                      help="Backend servers that are part of this pool. Format is colon deliminated. "
                           "BACKEND_IP:PORT:WEIGHT. eg. 10.0.0.1:80:50")
# https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_L7HealthMonitor/
@click.option('--healthPath', default='/', show_default=True, help="Health check path.")
@click.option('--healthInterval', default=5, type=int, show_default=True, help="Health check interval between checks.")
@click.option('--healthRetry', default=2, type=int, show_default=True,
              help="Health check number of times before marking as DOWN.")
@click.option('--healthTimeout', default=2, type=int, show_default=True, help="Health check timeout.")
# https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_L7SessionAffinity/
@click.option('--sticky', '-s', is_flag=True, callback=sticky_option, help="Make sessions sticky based on source_ip.")
@environment.pass_env
def l7pool_add(env, identifier, **args):
    """Adds a new l7 pool

    -S is in colon deliminated format to make grouping IP:port:weight a bit easier.

    Example::

        slcli loadbal l7pool-add (--id LOADBAL_ID) (-n, --name NAME) [-m, --method METHOD]
        [-s, --server BACKEND_IP:PORT] [-p, --protocol PROTOCOL] [--health-path PATH]
        [--health-interval INTERVAL] [--health-retry RETRY]
        [--health-timeout TIMEOUT] [--sticky cookie | source-ip]
    """

    mgr = SoftLayer.LoadBalancerManager(env.client)
    uuid, _ = mgr.get_lbaas_uuid_id(identifier)

    pool_main = {
        'name': args.get('name'),
        'loadBalancingAlgorithm': args.get('method'),
        'protocol': args.get('protocol')
    }

    pool_members = list(args.get('server'))

    pool_health = {
        'interval': args.get('healthinterval'),
        'timeout': args.get('healthtimeout'),
        'maxRetries': args.get('healthretry'),
        'urlPath': args.get('healthpath')
    }

    pool_sticky = {
        'type': args.get('sticky')
    }

    try:
        mgr.add_lb_l7_pool(uuid, pool_main, pool_members, pool_health, pool_sticky)
        click.secho("Success", fg='green')
    except SoftLayerAPIError as exception:
        click.secho(f"ERROR: {exception.faultString}", fg='red')


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def l7pool_del(env, identifier):
    """Deletes the identified pool

    Identifier is L7Pool Id. NOT the UUID
    """
    mgr = SoftLayer.LoadBalancerManager(env.client)
    try:
        mgr.del_lb_l7_pool(identifier)
        click.secho("Success", fg='green')
    except SoftLayerAPIError as exception:
        click.secho(f"ERROR: {exception.faultString}", fg='red')



"""
    SoftLayer.transports.rest
    ~~~~~~~~~~~~~~~~~~~~
    REST Style transport library

    :license: MIT, see LICENSE for more details.
"""

import json
import logging
import requests

from SoftLayer import consts
from SoftLayer import exceptions

from .transport import _format_object_mask
from .transport import _proxies_dict
from .transport import ComplexEncoder
from .transport import get_session
from .transport import SoftLayerListResult

REST_SPECIAL_METHODS = {
    # 'deleteObject': 'DELETE',
    'createObject': 'POST',
    'createObjects': 'POST',
    'editObject': 'PUT',
    'editObjects': 'PUT',
}


class RestTransport(object):
    """REST transport.

    REST calls should mostly work, but is not fully tested.
    XML-RPC should be used when in doubt
    """

    def __init__(self, endpoint_url=None, timeout=None, proxy=None, user_agent=None, verify=True):

        self.endpoint_url = (endpoint_url or consts.API_PUBLIC_ENDPOINT_REST).rstrip('/')
        self.timeout = timeout or None
        self.proxy = proxy
        self.user_agent = user_agent or consts.USER_AGENT
        self.verify = verify
        self._client = None
        self.logger = logging.getLogger(__name__)

    @property
    def client(self):
        """Returns client session object"""

        if self._client is None:
            self._client = get_session(self.user_agent)
        return self._client

    def __call__(self, request):
        """Makes a SoftLayer API call against the REST endpoint.

        REST calls should mostly work, but is not fully tested.
        XML-RPC should be used when in doubt

        :param request request: Request object
        """
        params = request.headers.copy()
        if request.mask:
            request.mask = _format_object_mask(request.mask)
            params['objectMask'] = request.mask

        if request.limit or request.offset:
            limit = request.limit or 0
            offset = request.offset or 0
            params['resultLimit'] = "%d,%d" % (offset, limit)

        if request.filter:
            params['objectFilter'] = json.dumps(request.filter)

        request.params = params

        # This handles any edge cases on the REST api.
        request.special_rest_params()

        auth = None
        if request.transport_user:
            auth = requests.auth.HTTPBasicAuth(
                request.transport_user,
                request.transport_password,
            )

        method = REST_SPECIAL_METHODS.get(request.method)

        if method is None:
            method = 'GET'

        body = {}
        if request.args:
            # NOTE(kmcdonald): force POST when there are arguments because
            # the request body is ignored otherwise.
            method = 'POST'
            body['parameters'] = request.args

        if body:
            request.payload = json.dumps(body, cls=ComplexEncoder)

        url_parts = [self.endpoint_url, request.service]
        if request.identifier is not None:
            url_parts.append(str(request.identifier))

        if request.method is not None:
            url_parts.append(request.method)

        request.url = '%s.%s' % ('/'.join(url_parts), 'json')

        # Prefer the request setting, if it's not None
        if request.verify is None:
            request.verify = self.verify
        try:
            resp = self.client.request(method, request.url,
                                       auth=auth,
                                       headers=request.transport_headers,
                                       params=request.params,
                                       data=request.payload,
                                       timeout=self.timeout,
                                       verify=request.verify,
                                       cert=request.cert,
                                       proxies=_proxies_dict(self.proxy))

            request.url = resp.url

            resp.raise_for_status()

            if resp.text != "":
                try:
                    result = json.loads(resp.text)
                except ValueError as json_ex:
                    self.logger.warning(json_ex)
                    raise exceptions.SoftLayerAPIError(resp.status_code, str(resp.text))
            else:
                raise exceptions.SoftLayerAPIError(resp.status_code, "Empty response.")

            request.result = result

            if isinstance(result, list):
                return SoftLayerListResult(result, int(resp.headers.get('softlayer-total-items', 0)))
            else:
                return result
        except requests.HTTPError as ex:
            try:
                message = json.loads(ex.response.text)['error']
                request.url = ex.response.url
            except ValueError as json_ex:
                if ex.response.text == "":
                    raise exceptions.SoftLayerAPIError(resp.status_code, "Empty response.")
                self.logger.warning(json_ex)
                raise exceptions.SoftLayerAPIError(resp.status_code, ex.response.text)

            raise exceptions.SoftLayerAPIError(ex.response.status_code, message)
        except requests.RequestException as ex:
            raise exceptions.TransportError(0, str(ex))

    @staticmethod
    def print_reproduceable(request):
        """Prints out the minimal python code to reproduce a specific request

        The will also automatically replace the API key so its not accidently exposed.

        :param request request: Request object
        """
        # This handles any edge cases on the REST api.
        request.special_rest_params()
        command = "curl -u $SL_USER:$SL_APIKEY -X {method} -H {headers} {data} '{uri}'"

        method = REST_SPECIAL_METHODS.get(request.method)

        if method is None:
            method = 'GET'
        if request.args:
            method = 'POST'

        data = ''
        if request.payload is not None:
            data = f"-d '{request.payload}'"

        headers = ['"{0}: {1}"'.format(k, v) for k, v in request.transport_headers.items()]
        headers = " -H ".join(headers)
        return command.format(method=method, headers=headers, data=data, uri=request.url)



"""Cancel global IP."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('-f', '--force', default=False, is_flag=True, help="Force operation without confirmation")
@environment.pass_env
def cli(env, identifier, force):
    """Cancel global IP.

    Example::

        slcli globalip cancel 12345
    """

    mgr = SoftLayer.NetworkManager(env.client)
    global_ip_id = helpers.resolve_id(mgr.resolve_global_ip_ids, identifier, name='global ip')

    if not force:
        if not (env.skip_confirmations or
                formatting.confirm(f"This will cancel the IP address: {global_ip_id} and cannot be undone. Continue?")):
            raise exceptions.CLIAbort('Aborted')

    mgr.cancel_global_ip(global_ip_id)



"""Print zone in BIND format."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('zone')
@environment.pass_env
def cli(env, zone):
    """Print zone in BIND format.

    Example::
        slcli dns zone-print ibm.com
        This command prints zone that is named ibm.com, and in BIND format.
"""

    manager = SoftLayer.DNSManager(env.client)
    zone_id = helpers.resolve_id(manager.resolve_ids, zone, name='zone')
    env.fout(manager.dump_zone(zone_id))



"""List Tags."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.exceptions import SoftLayerAPIError
from SoftLayer.managers.tags import TagManager
from SoftLayer import utils

# pylint: disable=unnecessary-lambda


@click.command(cls=SLCommand)
@click.option('--detail', '-d', is_flag=True, default=False,
              help="Show information about the resources using this tag.")
@environment.pass_env
def cli(env, detail):
    """List Tags."""

    tag_manager = TagManager(env.client)

    if detail:
        tables = detailed_table(tag_manager, tag_manager.get_attached_tags())
        for table in tables:
            env.fout(table)
    else:
        table = simple_table(tag_manager)
        env.fout(table)
    # pp(tags.list_tags())


def tag_row(tag):
    """Format a tag table row"""
    return [tag.get('id'), tag.get('name'), tag.get('referenceCount', 0)]


def detailed_table(tag_manager, tags):
    """Creates a table for each tag, with details about resources using it"""
    tables = []
    for tag in tags:
        references = tag_manager.get_tag_references(tag.get('id'))
        # pp(references)
        new_table = formatting.Table(['Id', 'Type', 'Resource'], title=tag.get('name'))
        for reference in references:
            tag_type = utils.lookup(reference, 'tagType', 'keyName')
            resource_id = reference.get('resourceTableId')
            resource_row = get_resource_name(tag_manager, resource_id, tag_type)
            new_table.add_row([resource_id, tag_type, resource_row])
        tables.append(new_table)

    return tables


def simple_table(tag_manager):
    """Just tags and how many resources on each"""
    tags = tag_manager.list_tags()
    table = formatting.Table(['Id', 'Tag', 'Count'], title='Tags')
    for tag in tags.get('attached', []):
        table.add_row(tag_row(tag))
    for tag in tags.get('unattached', []):
        table.add_row(tag_row(tag))
    return table


def get_resource_name(tag_manager, resource_id, tag_type):
    """Returns a string to identify a resource"""
    name = None
    try:
        resource = tag_manager.reference_lookup(resource_id, tag_type)
        name = tag_manager.get_resource_name(resource, tag_type)
    except SoftLayerAPIError as exception:
        name = f"{exception.reason}"
    return name



"""Bandwidth summary for every pool/server."""
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SLCommand, short_help="Bandwidth summary for every pool/server")
@environment.pass_env
def cli(env):
    """Bandwidth summary for every pool/server.

    This summary on the total data transfered for each virtual sever, hardware
    server and bandwidth pool.
    https://cloud.ibm.com/classic-bandwidth

    More information
    https://cloud.ibm.com/docs/bandwidth-metering?topic=bandwidth-metering-get-started-with-bandwidth-metering
    """

    table = formatting.Table([
        'Id',
        'Device name',
        'Location',
        'Allocation',
        'Data in',
        'Data out',
        'Total usage',
        'Pool',
        'Tags',
    ])

    mask = """mask[resource(SoftLayer_Hardware)[id,bandwidthAllocation,bandwidthAllotmentDetail[id,bandwidthAllotment
    [id,bandwidthAllotmentTypeId,name]],billingItem[id,createDate,lastBillDate],datacenter[id,name],
    fullyQualifiedDomainName,inboundPublicBandwidthUsage,outboundPublicBandwidthUsage,primaryIpAddress,tagReferences
    [id,tag[id,name]]],resource(SoftLayer_Network_Application_Delivery_Controller)[id,billingItem[id,
    bandwidthAllocation[id,amount],bandwidthAllotmentDetail[id,bandwidthAllotment[id,bandwidthAllotmentTypeId,name]]
    ,createDate,lastBillDate],datacenter[id,name],name,outboundPublicBandwidthUsage,primaryIpAddress,tagReferences
    [id,tag[id,name]]],resource(SoftLayer_Virtual_Guest)[id,bandwidthAllocation,bandwidthAllotmentDetail[id,
    bandwidthAllotment[id,bandwidthAllotmentTypeId,name]],billingItem[id,createDate,lastBillDate],datacenter[id,name
    ],fullyQualifiedDomainName,inboundPublicBandwidthUsage,outboundPublicBandwidthUsage,primaryIpAddress,
    tagReferences[id,tag[id,name]]]]"""

    search_string = """_objectType:SoftLayer_Hardware,SoftLayer_Virtual_Guest,
    SoftLayer_Network_Application_Delivery_Controller _sort:[fullyQualifiedDomainName:asc]"""

    servers = env.client.call(
        'Search', 'advancedSearch',
        search_string,
        mask=mask,
        iter=True
    )

    for server in servers:
        resource = server.get('resource')

        device_name = utils.lookup(resource, 'fullyQualifiedDomainName')
        if not device_name:
            device_name = utils.lookup(resource, 'name')

        bandwidth_allocation = utils.lookup(resource, 'bandwidthAllocation')
        if bandwidth_allocation != '0':
            if bandwidth_allocation is not None:
                bandwidth_allocation = formatting.convert_sizes(
                    bandwidth_allocation, round_result=True)
            else:
                bandwidth_allocation = 'Unlimited'
        else:
            bandwidth_allocation = 'Pay-As-You-Go'

        in_bandwidth_public = formatting.convert_sizes(utils.lookup(resource, 'inboundPublicBandwidthUsage'))

        out_bandwidth_public = formatting.convert_sizes(utils.lookup(resource, 'outboundPublicBandwidthUsage'))

        total_bandwidth_public = formatting.sum_sizes(in_bandwidth_public, out_bandwidth_public)

        if bandwidth_allocation != 'Unlimited' and bandwidth_allocation != 'Pay-As-You-Go':
            pool = utils.lookup(resource, 'bandwidthAllotmentDetail', 'bandwidthAllotment', 'name')
        else:
            pool = 'Not Applicable'

        table.add_row([
            utils.lookup(resource, 'id'),
            device_name,
            utils.lookup(resource, 'datacenter', 'name'),
            bandwidth_allocation,
            in_bandwidth_public,
            out_bandwidth_public,
            total_bandwidth_public,
            pool,
            formatting.tags(resource.get('tagReferences')),
        ])

    env.fout(table)



"""
    SoftLayer.account
    ~~~~~~~~~~~~~~~~~~~~~~~
    Account manager

    :license: MIT, see License for more details.
"""

import logging

from SoftLayer.exceptions import SoftLayerAPIError
from SoftLayer import utils

# Invalid names are ignored due to long method names and short argument names
# pylint: disable=invalid-name

LOGGER = logging.getLogger(__name__)


class AccountManager(utils.IdentifierMixin, object):
    """Common functions for getting information from the Account service

    :param SoftLayer.API.BaseClient client: the client instance
    """
    _DEFAULT_BILLING_ITEM_MASK = """mask[
                    orderItem[id,order[id,userRecord[id,email,displayName,userStatus]]],
                    nextInvoiceTotalRecurringAmount,
                    location, hourlyFlag, children
                ]"""

    def __init__(self, client):
        self.client = client

    def get_summary(self):
        """Gets some basic account information

        :return: Account object
        """
        mask = """mask[
            nextInvoiceTotalAmount,
            pendingInvoice[invoiceTotalAmount],
            blockDeviceTemplateGroupCount,
            dedicatedHostCount,
            domainCount,
            hardwareCount,
            networkStorageCount,
            openTicketCount,
            networkVlanCount,
            subnetCount,
            userCount,
            virtualGuestCount
            ]
        """
        return self.client.call('Account', 'getObject', mask=mask)

    def get_upcoming_events(self, event_type, date_min=None):
        """Retrieves a list of Notification_Occurrence_Events that have not ended yet

        :param: String event_type: notification event type.
        :param: String date_min: greater Than Date to data recovery, default is 2 days ago.
        :return: SoftLayer_Notification_Occurrence_Event
        """
        mask = "mask[id, subject, startDate, endDate, modifyDate, statusCode, acknowledgedFlag, " \
               "impactedResourceCount, updateCount, systemTicketId, notificationOccurrenceEventType[keyName]]"

        _filter = {
            'notificationOccurrenceEventType': {
                'keyName': {
                    'operation': event_type
                }
            }
        }

        self.add_event_filter(_filter, event_type, date_min)

        return self.client.call('Notification_Occurrence_Event', 'getAllObjects', filter=_filter, mask=mask, iter=True)

    @staticmethod
    def add_event_filter(_filter, event_type, date_min=None):
        """Add data to the object filter.

        :param: _filter: event filter.
        :param: string event_type: event type.
        :param: string date_min: greater Than Date to data recovery, default is 2 days ago.
        """

        if event_type == 'PLANNED':
            if date_min:
                _filter['endDate'] = {
                    'operation': 'greaterThanDate',
                    'options': [{
                        'name': 'date',
                        'value': [date_min]
                    }]
                }
            else:
                _filter['endDate'] = {
                    'operation': '> sysdate - 2'
                }
            _filter['startDate'] = {
                'operation': 'orderBy',
                'options': [{
                    'name': 'sort',
                    'value': ['DESC']
                }]
            }

        if event_type == 'UNPLANNED_INCIDENT':
            if date_min:
                _filter['modifyDate'] = {
                    'operation': 'greaterThanDate',
                    'options': [{
                        'name': 'date',
                        'value': [date_min]
                    }]
                }
            else:
                _filter['modifyDate'] = {
                    'operation': '> sysdate - 2'
                }

        if event_type == 'ANNOUNCEMENT':
            _filter['statusCode'] = {
                'keyName': {
                    'operation': 'in',
                    'options': [{
                        'name': 'data',
                        'value': ['PUBLISHED']
                    }]
                }
            }

    def ack_event(self, event_id):
        """Acknowledge an event. This mostly prevents it from appearing as a notification in the control portal.

        :param int event_id: Notification_Occurrence_Event ID you want to ack
        :return: True on success, Exception otherwise.
        """
        return self.client.call('Notification_Occurrence_Event', 'acknowledgeNotification', id=event_id)

    def get_event(self, event_id):
        """Gets details about a maintenance event

        :param int event_id: Notification_Occurrence_Event ID
        :return: Notification_Occurrence_Event
        """
        mask = """mask[
            acknowledgedFlag,
            attachments,
            impactedResources,
            statusCode,
            updates,
            notificationOccurrenceEventType]
        """
        return self.client.call('Notification_Occurrence_Event', 'getObject', id=event_id, mask=mask)

    def get_invoices(self, limit=50, closed=False, get_all=False):
        """Gets an accounts invoices.

        :param int limit: Number of invoices to get back in a single call.
        :param bool closed: If True, will also get CLOSED invoices
        :param bool get_all: If True, will paginate through invoices until all have been retrieved.
        :return: Billing_Invoice
        """
        mask = "mask[invoiceTotalAmount, itemCount]"
        _filter = {
            'invoices': {
                'createDate': {
                    'operation': 'orderBy',
                    'options': [{
                        'name': 'sort',
                        'value': ['DESC']
                    }]
                },
                'statusCode': {'operation': 'OPEN'},
            }
        }
        if closed:
            del _filter['invoices']['statusCode']

        return self.client.call('Account', 'getInvoices', mask=mask, filter=_filter, iter=get_all, limit=limit)

    def get_billing_items(self, identifier):
        """Gets all topLevelBillingItems from a specific invoice

        :param int identifier: Invoice Id
        :return: Billing_Invoice_Item
        """

        mask = """mask[
            id, description, hostName, domainName, oneTimeAfterTaxAmount, recurringAfterTaxAmount, createDate,
            categoryCode,
            category[name],
            location[name],
            children[id, category[name], description, oneTimeAfterTaxAmount, recurringAfterTaxAmount]
        ]"""
        return self.client.call(
            'Billing_Invoice',
            'getInvoiceTopLevelItems',
            id=identifier,
            mask=mask,
            iter=True,
            limit=100
        )

    def get_account_billing_items(self, create=None, category=None, mask=None):
        """Gets all the topLevelBillingItems currently active on the account

        :param string mask: Object Mask
        :return: Billing_Item
        """

        if mask is None:
            mask = """mask[
                orderItem[id,order[id,userRecord[id,email,displayName,userStatus]]],
                nextInvoiceTotalRecurringAmount,
                location, hourlyFlag
            ]"""

        object_filter = {
            "allTopLevelBillingItems": {
                "cancellationDate": {
                    "operation": "is null"
                },
                "id": {
                    "operation": "orderBy",
                    "options": [
                        {
                            "name": "sort",
                            "value": ["ASC"]
                        }
                    ]
                }
            }
        }

        if category:
            object_filter = utils.dict_merge(object_filter,
                                             {"allTopLevelBillingItems": {"categoryCode": {"operation": category}}})
        if create:
            object_filter = utils.dict_merge(object_filter,
                                             {"allTopLevelBillingItems": {"createDate": {"operation": "*=" + create}}})

        return self.client.call('Account', 'getAllTopLevelBillingItems',
                                mask=mask, filter=object_filter, iter=True, limit=100)

    def get_billing_item(self, identifier, mask=None):
        """Gets details about a billing item

        :param int identifier: Billing_Item id
        :param string mask: Object mask to use.
        :return: Billing_Item
        """

        if mask is None:
            mask = self._DEFAULT_BILLING_ITEM_MASK

        return self.client.call('Billing_Item', 'getObject', id=identifier, mask=mask)

    def get_billing_item_from_invoice(self, identifier, mask=None):
        """Gets details about a billing item of a billing invoice item

        :param int identifier: Billing_Invoice_Item id
        :param mask: Object mask to use.
        :return: Billing_Item
        """
        if mask is None:
            mask = self._DEFAULT_BILLING_ITEM_MASK
        return self.client.call('Billing_Invoice_Item', 'getBillingItem', id=identifier, mask=mask)

    def get_item_detail(self, identifier):
        """Gets details about a billing item

        :param int identifier: Billing_Item id or Billing_Invoice_Item
        :return: Billing_Item
        """

        try:
            return self.get_billing_item(identifier)
        except SoftLayerAPIError as exception:
            if exception.faultCode == 404:
                return self.get_billing_item_from_invoice(identifier)
            raise

    def cancel_item(self, identifier, reason="No longer needed", note=None):
        """Cancels a specific billing item with a reason

        :param int identifier: Billing_Item id
        :param string reason: A cancellation reason
        :param string note: Custom note to set when cancelling. Defaults to information about who canceled the item.
        :return: bool
        """

        if note is None:
            user = self.client.call('Account', 'getCurrentUser', mask="mask[id,displayName,email,username]")
            note = f"Cancelled by {user.get('username')} with the SLCLI"

        return self.client.call('Billing_Item', 'cancelItem', False, True, reason, note, id=identifier)

    def get_account_all_billing_orders(self, limit=100, mask=None):
        """Gets all the topLevelBillingItems currently active on the account

        :param string mask: Object Mask
        :return: Billing_Item
        """

        if mask is None:
            mask = """
                  orderTotalAmount, userRecord,
                  initialInvoice[id,amount,invoiceTotalAmount],
                  items[description]
               """
        return self.client.call('Billing_Order', 'getAllObjects',
                                limit=limit, mask=mask)

    def get_routers(self, location=None, mask=None):
        """Gets all the routers currently active on the account

        :param string mask: Object Mask
        :param string location: location string
        :returns: Routers
        """

        if mask is None:
            mask = """
                 topLevelLocation
               """
        object_filter = ''
        if location:
            object_filter = {
                'routers': {
                    'topLevelLocation': {'name': {'operation': location}}
                }
            }

        return self.client['SoftLayer_Account'].getRouters(filter=object_filter, mask=mask)

    def get_network_message_delivery_accounts(self):
        """Gets all Network Message delivery accounts.

        :returns: Network Message delivery accounts
        """

        _mask = """vendor,type"""

        return self.client['SoftLayer_Account'].getNetworkMessageDeliveryAccounts(mask=_mask)

    def get_active_virtual_licenses(self):
        """Gets all active virtual licenses account.

        :returns: active virtual licenses account
        """

        _mask = """billingItem[categoryCode,createDate,description],
                    key,id,ipAddress,
                    softwareDescription[longDescription,name,manufacturer],
                    subnet"""

        return self.client['SoftLayer_Account'].getActiveVirtualLicenses(mask=_mask)

    def get_active_account_licenses(self):
        """Gets all active account licenses.

        :returns: Active account Licenses
        """

        _mask = """billingItem,softwareDescription"""

        return self.client['SoftLayer_Account'].getActiveAccountLicenses(mask=_mask)

    def get_bandwidth_pools(self, mask=None):
        """Gets all the bandwidth pools on an account"""

        if mask is None:
            mask = """mask[totalBandwidthAllocated,locationGroup, id, name, projectedPublicBandwidthUsage,
                      billingCyclePublicBandwidthUsage[amountOut,amountIn],
                      billingItem[id,nextInvoiceTotalRecurringAmount],outboundPublicBandwidthUsage,serviceProviderId,bandwidthAllotmentTypeId,activeDetailCount,endDate]
                   """

        return self.client.call('SoftLayer_Account', 'getBandwidthAllotments', mask=mask, iter=True)

    def get_bandwidth_pool_counts(self, identifier):
        """Gets a count of all servers in a bandwidth pool

        Getting the server counts individually is significantly faster than pulling them in
        with the get_bandwidth_pools api call.
        """
        mask = "mask[id, bareMetalInstanceCount, hardwareCount, virtualGuestCount]"
        counts = self.client.call('SoftLayer_Network_Bandwidth_Version1_Allotment', 'getObject',
                                  id=identifier, mask=mask)
        total = counts.get('bareMetalInstanceCount', 0) + \
            counts.get('hardwareCount', 0) + \
            counts.get('virtualGuestCount', 0)
        return total

    def getBandwidthDetail(self, identifier):
        """Gets bandwidth pool detail.

        :returns: bandwidth pool detail
        """
        _mask = """activeDetails[allocation],projectedPublicBandwidthUsage, billingCyclePublicBandwidthUsage,
        hardware[outboundBandwidthUsage,bandwidthAllotmentDetail[allocation]],inboundPublicBandwidthUsage,
        virtualGuests[outboundPublicBandwidthUsage,bandwidthAllotmentDetail[allocation]],
        bareMetalInstances[outboundBandwidthUsage,bandwidthAllotmentDetail[allocation]]"""
        return self.client['SoftLayer_Network_Bandwidth_Version1_Allotment'].getObject(id=identifier, mask=_mask)

    def get_provisioning_scripts(self):
        """Gets a provisioning hooks.

        :returns: provisioning hook
        """

        return self.client.call('Account', 'getPostProvisioningHooks')

    def create_provisioning(self, name, uri):
        """create a provisioning script

        :param name: Name of the hook.
        :param uri: endpoint that the script will be downloaded
        """
        template = {
            'name': name,
            'uri': uri
        }
        return self.client.call('SoftLayer_Provisioning_Hook', 'createObject', template)

    def delete_provisioning(self, identifier):
        """Delete a provisioning script

        param: identifier provisioning script identifier

        Returns: boolean
        """
        return self.client.call("SoftLayer_Provisioning_Hook", "deleteObject", id=identifier)

    def get_account_upgrade_orders(self, limit=100):
        """Gets upgrade order list"""
        return self.client.call('SoftLayer_Account', 'getUpgradeRequests', limit=limit)



getAllObjects = [
    {'description': 'Email about your order.',
     'enabled': True,
     'id': 1,
     'name': 'Order Being Reviewed'
     },
    {'description': 'Maintenances that will or are likely to cause service '
                    'outages and disruptions',
     'enabled': True,
     'id': 8,
     'name': 'High Impact'
     },
    {'description': 'Testing description.',
     'enabled': True,
     'id': 111,
     'name': 'Test notification'
     }
]

enable = True

disable = True



"""Cancel virtual servers."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Cancel virtual servers."""

    vsi = SoftLayer.VSManager(env.client)
    vs_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')
    if not (env.skip_confirmations or formatting.no_going_back(vs_id)):
        raise exceptions.CLIAbort('Aborted')

    virtual_server = vsi.cancel_instance(vs_id)

    if virtual_server:
        env.fout(f"The virtual server instance: {vs_id} was cancelled.")



"""Enable or Disable vpn for a user."""
# :license: MIT, see LICENSE for more details.


import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('user')
@environment.pass_env
def vpn_enable(env, user):
    """Enable vpn for a user.

    Example::
        slcli user vpn-enable 1234567
    """

    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, user, 'username')

    result = mgr.vpn_enable_or_disable(user_id, True)
    message = f"{user} vpn is successfully enabled"

    if result:
        click.secho(message, fg='green')


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('user')
@environment.pass_env
def vpn_disable(env, user):
    """Disable vpn for a user.

    Example::
        slcli user vpn-disable 1234567
    """

    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, user, 'username')

    result = mgr.vpn_enable_or_disable(user_id, False)
    message = f"{user} vpn is successfully disabled"

    if result:
        click.secho(message, fg='green')



"""
    SoftLayer.vs_capacity
    ~~~~~~~~~~~~~~~~~~~~~~~
    Reserved Capacity Manager and helpers

    :license: MIT, see License for more details.
"""

import logging

from SoftLayer.exceptions import SoftLayerError
from SoftLayer.managers import ordering
from SoftLayer.managers.vs import VSManager
from SoftLayer import utils

# Invalid names are ignored due to long method names and short argument names
# pylint: disable=invalid-name

LOGGER = logging.getLogger(__name__)


class CapacityManager(utils.IdentifierMixin, object):
    """Manages SoftLayer Reserved Capacity Groups.

        Product Information

        - https://cloud.ibm.com/docs/virtual-servers?topic=virtual-servers-about-reserved-virtual-servers
        - https://softlayer.github.io/reference/services/SoftLayer_Virtual_ReservedCapacityGroup/
        - https://softlayer.github.io/reference/services/SoftLayer_Virtual_ReservedCapacityGroup_Instance/


    :param SoftLayer.API.BaseClient client: the client instance
    :param SoftLayer.managers.OrderingManager ordering_manager: an optional manager to handle ordering.
                                              If none is provided, one will be auto initialized.
    """

    def __init__(self, client, ordering_manager=None):
        self.client = client
        self.account = client['Account']
        self.capacity_package = 'RESERVED_CAPACITY'
        self.rcg_service = 'Virtual_ReservedCapacityGroup'

        if ordering_manager is None:
            self.ordering_manager = ordering.OrderingManager(client)

    def list(self):
        """List Reserved Capacities"""
        mask = """mask[availableInstanceCount, occupiedInstanceCount,
instances[id, billingItem[description, hourlyRecurringFee]], instanceCount, backendRouter[datacenter]]"""
        results = self.client.call('Account', 'getReservedCapacityGroups', mask=mask)
        return results

    def get_object(self, identifier, mask=None):
        """Get a Reserved Capacity Group

        :param int identifier: Id of the SoftLayer_Virtual_ReservedCapacityGroup
        :param string mask: override default object Mask
        """
        if mask is None:
            mask = "mask[instances[billingItem[item[keyName],category], guest], backendRouter[datacenter]]"
        result = self.client.call(self.rcg_service, 'getObject', id=identifier, mask=mask)
        return result

    def get_create_options(self):
        """List available reserved capacity plans"""
        mask = "mask[attributes,prices[pricingLocationGroup]]"
        results = self.ordering_manager.list_items(self.capacity_package, mask=mask)
        return results

    def get_available_routers(self, dc=None):
        """Pulls down all backendRouterIds that are available

        :param string dc: A specific location to get routers for, like 'dal13'.
        :returns list: A list of locations where RESERVED_CAPACITY can be ordered.
        """
        mask = "mask[locations]"
        # Step 1, get the package id
        package = self.ordering_manager.get_package_by_key(self.capacity_package, mask="id")

        # Step 2, get the regions this package is orderable in
        regions = self.client.call('Product_Package', 'getRegions', id=package['id'], mask=mask, iter=True)
        _filter = None
        routers = {}
        if dc is not None:
            _filter = {'datacenterName': {'operation': dc}}

        # Step 3, for each location in each region, get the pod details, which contains the router id
        pods = self.client.call('Network_Pod', 'getAllObjects', filter=_filter, iter=True)
        for region in regions:
            routers[region['keyname']] = []
            for location in region['locations']:
                location['location']['pods'] = []
                for pod in pods:
                    if pod['datacenterName'] == location['location']['name']:
                        location['location']['pods'].append(pod)

        # Step 4, return the data.
        return regions

    def create(self, name, backend_router_id, flavor, instances, test=False):
        """Orders a Virtual_ReservedCapacityGroup

        :param string name: Name for the new reserved capacity
        :param int backend_router_id: This selects the pod. See create_options for a list
        :param string flavor: Capacity KeyName, see create_options for a list
        :param int instances: Number of guest this capacity can support
        :param bool test: If True, don't actually order, just test.
        """

        # Since orderManger needs a DC id, just send in 0, the API will ignore it
        args = (self.capacity_package, 0, [flavor])
        extras = {"backendRouterId": backend_router_id, "name": name}
        kwargs = {
            'extras': extras,
            'quantity': instances,
            'complex_type': 'SoftLayer_Container_Product_Order_Virtual_ReservedCapacity',
            'hourly': True
        }
        if test:
            receipt = self.ordering_manager.verify_order(*args, **kwargs)
        else:
            receipt = self.ordering_manager.place_order(*args, **kwargs)
        return receipt

    def create_guest(self, capacity_id, test, guest_object):
        """Turns an empty Reserve Capacity into a real Virtual Guest

        :param int capacity_id: ID of the RESERVED_CAPACITY_GROUP to create this guest into
        :param bool test: True will use verifyOrder, False will use placeOrder
        :param dictionary guest_object:  Below is the minimum info you need to send in
            guest_object = {
            'domain': 'test.com',
            'hostname': 'A1538172419',
            'os_code': 'UBUNTU_LATEST_64',
            'primary_disk': '25',
            }

        """

        vs_manager = VSManager(self.client)
        mask = "mask[instances[id, billingItem[id, item[id,keyName]]], backendRouter[id, datacenter[name]]]"
        capacity = self.get_object(capacity_id, mask=mask)
        try:
            capacity_flavor = capacity['instances'][0]['billingItem']['item']['keyName']
            flavor = _flavor_string(capacity_flavor, guest_object['primary_disk'])
        except KeyError as ex:
            raise SoftLayerError("Unable to find capacity Flavor.") from ex

        guest_object['flavor'] = flavor
        guest_object['datacenter'] = capacity['backendRouter']['datacenter']['name']

        # Reserved capacity only supports SAN as of 20181008
        guest_object['local_disk'] = False
        # Reserved capacity only supports monthly ordering via Virtual_Guest::generateOrderTemplate
        # Hourly ordering would require building out the order manually.
        guest_object['hourly'] = False
        template = vs_manager.verify_create_instance(**guest_object)
        template['reservedCapacityId'] = capacity_id
        if guest_object.get('ipv6'):
            ipv6_price = self.ordering_manager.get_price_id_list('PUBLIC_CLOUD_SERVER', ['1_IPV6_ADDRESS'])
            template['prices'].append({'id': ipv6_price[0]})

        if test:
            result = self.client.call('Product_Order', 'verifyOrder', template)
        else:
            result = self.client.call('Product_Order', 'placeOrder', template)

        return result


def _flavor_string(capacity_key, primary_disk):
    """Removed the _X_YEAR_TERM from capacity_key and adds the primary disk size, creating the flavor keyName

    This will work fine unless 10 year terms are invented... or flavor format changes...
    """
    flavor = "%sX%s" % (capacity_key[:-12], primary_disk)
    return flavor



"""List subnets."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--sortby',
              help='Column to sort by',
              type=click.Choice(['id',
                                 'identifier',
                                 'type',
                                 'network_space',
                                 'datacenter',
                                 'vlan_id',
                                 'IPs',
                                 'hardware',
                                 'vs']))
@click.option('--datacenter', '-d',
              help="Filter by datacenter shortname (sng01, dal05, ...)")
@click.option('--identifier', help="Filter by network identifier")
@click.option('--subnet-type', '-t', help="Filter by subnet type")
@click.option('--network-space', help="Filter by network space")
@click.option('--ipv4', '--v4', is_flag=True, help="Display only IPv4 subnets")
@click.option('--ipv6', '--v6', is_flag=True, help="Display only IPv6 subnets")
@environment.pass_env
def cli(env, sortby, datacenter, identifier, subnet_type, network_space, ipv4, ipv6):
    """List subnets."""

    mgr = SoftLayer.NetworkManager(env.client)

    table = formatting.Table([
        'Id', 'Identifier', 'Network', 'Type', 'VLAN', 'Location', 'Target',
        'IPs', 'Hardware', 'Vs', 'Tags', 'Note'
    ])
    table.sortby = sortby

    version = 0
    if ipv4:
        version = 4
    elif ipv6:
        version = 6

    subnets = mgr.list_subnets(
        datacenter=datacenter,
        version=version,
        identifier=identifier,
        subnet_type=subnet_type,
        network_space=network_space,
    )

    for subnet in subnets:
        subnet_type = subnet.get('subnetType', formatting.blank())
        if subnet_type == 'PRIMARY' or subnet_type == 'PRIMARY_6' or subnet_type == 'ADDITIONAL_PRIMARY':
            subnet_type = 'Primary'
        if subnet_type == 'SECONDARY_ON_VLAN':
            subnet_type = 'Portable'
        if subnet_type == 'STATIC_IP_ROUTED':
            subnet_type = 'Static'
        if subnet_type == 'GLOBAL_IP':
            subnet_type = 'Global'

        network = subnet.get('addressSpace', formatting.blank())
        if network is not None:
            network = str(network).capitalize()

        vlan = utils.lookup(subnet, 'networkVlan', 'fullyQualifiedName')
        if vlan is None:
            vlan = 'Unrouted'
        table.add_row([
            subnet['id'],
            '%s/%s' % (subnet['networkIdentifier'], str(subnet['cidr'])),
            network,
            subnet_type,
            vlan,
            utils.lookup(subnet, 'datacenter', 'name') or formatting.blank(),
            utils.lookup(subnet, 'endPointIpAddress', 'ipAddress') or formatting.blank(),
            subnet['ipAddressCount'],
            len(subnet['hardware']),
            len(subnet['virtualGuests']),
            formatting.tags(subnet.get('tagReferences')),
            utils.lookup(subnet, 'note') or formatting.blank(),
        ])

    env.fout(table)



"""
    SoftLayer.user
    ~~~~~~~~~~~~~
    User Manager/helpers

    :license: MIT, see LICENSE for more details.
"""
import datetime
import logging
from operator import itemgetter

from SoftLayer import exceptions
from SoftLayer import utils

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-public-methods
class UserManager(utils.IdentifierMixin, object):
    """Manages Users.

    See: https://softlayer.github.io/reference/datatypes/SoftLayer_User_Customer/

    Example::

       # Initialize the Manager.
       import SoftLayer
       client = SoftLayer.create_client_from_env()
       mgr = SoftLayer.UserManager(client)

    :param SoftLayer.API.BaseClient client: the client instance

    """

    def __init__(self, client):
        self.client = client
        self.user_service = self.client['SoftLayer_User_Customer']
        self.override_service = self.client['Network_Service_Vpn_Overrides']
        self.account_service = self.client['SoftLayer_Account']
        self.subscription_service = self.client['SoftLayer_Email_Subscription']
        self.resolvers = [self._get_id_from_username]
        self.all_permissions = None

    def list_users(self, objectmask=None, objectfilter=None):
        """Lists all users on an account

        :param string objectmask: Used to overwrite the default objectmask.
        :param dictionary objectfilter: If you want to use an objectfilter.
        :returns: A list of dictionaries that describe each user

        Example::
            result = mgr.list_users()
        """

        if objectmask is None:
            objectmask = """mask[id, username, displayName, userStatus[name], hardwareCount, virtualGuestCount,
                                 email, roles, externalBindingCount,apiAuthenticationKeyCount, sslVpnAllowedFlag]"""

        return self.account_service.getUsers(mask=objectmask, filter=objectfilter)

    def get_user(self, user_id, objectmask=None):
        """Calls SoftLayer_User_Customer::getObject

        :param int user_id: Id of the user
        :param string objectmask: default is 'mask[userStatus[name], parent[id, username]]'
        :returns: A user object.
        """
        if objectmask is None:
            objectmask = "mask[userStatus[name], parent[id, username]]"
        return self.user_service.getObject(id=user_id, mask=objectmask)

    def get_current_user(self, objectmask=None):
        """Calls SoftLayer_Account::getCurrentUser"""

        if objectmask is None:
            objectmask = "mask[userStatus[name], parent[id, username]]"
        return self.account_service.getCurrentUser(mask=objectmask)

    def get_all_permissions(self):
        """Calls User_Permission_Action::getAllObjects

        Stores the result in self.all_permissions
        :returns: A list of dictionaries that contains all valid permissions
        """
        if self.all_permissions is None:
            permissions = self.client.call('User_Permission_Action', 'getAllObjects')
            self.all_permissions = sorted(permissions, key=itemgetter('keyName'))
        return self.all_permissions

    def get_permission_departments(self):
        """Calls SoftLayer_User_Permission_Department::getAllObjects

        Stores the result in self.all_permissions
        :returns: A list of dictionaries that contains all valid permissions
        """
        mask = "mask[permissions[id,keyName,description,name]]"
        departments = self.client.call('User_Permission_Department', 'getAllObjects', mask=mask)
        for i, department in enumerate(departments):
            departments[i]['permissions'] = sorted(department.get('permissions'), key=itemgetter('keyName'))

        return departments

    def get_all_notifications(self):
        """Calls SoftLayer_Email_Subscription::getAllObjects

        Stores the result in self.all_permissions
        :returns: A list of dictionaries that contains all valid permissions
        """
        return self.subscription_service.getAllObjects(mask='mask[enabled]')

    def enable_notifications(self, notifications_names):
        """Enables a list of notifications for the current a user profile.

        :param list notifications_names: List of notifications names to enable
        :returns: True on success

        Example::

            enable_notifications(['Order Approved','Reload Complete'])
        """

        result = False
        notifications = self.gather_notifications(notifications_names)
        for notification in notifications:
            notification_id = notification.get('id')
            result = self.subscription_service.enable(id=notification_id)
            if not result:
                return False
        return result

    def disable_notifications(self, notifications_names):
        """Disable a list of notifications for the current a user profile.

        :param list notifications_names: List of notifications names to disable
        :returns: True on success

        Example::

            disable_notifications(['Order Approved','Reload Complete'])
        """

        result = False
        notifications = self.gather_notifications(notifications_names)
        for notification in notifications:
            notification_id = notification.get('id')
            result = self.subscription_service.disable(id=notification_id)
            if not result:
                return False
        return result

    def add_permissions(self, user_id, permissions):
        """Enables a list of permissions for a user

        :param int id: user id to set
        :param list permissions: List of permissions keynames to enable
        :returns: True on success, Exception otherwise

        Example::

            add_permissions(123, ['BANDWIDTH_MANAGE'])
        """
        pretty_permissions = self.format_permission_object(permissions)
        LOGGER.warning("Adding the following permissions to %s: %s", user_id, pretty_permissions)
        return self.user_service.addBulkPortalPermission(pretty_permissions, id=user_id)

    def remove_permissions(self, user_id, permissions):
        """Disables a list of permissions for a user

        :param int id: user id to set
        :param list permissions: List of permissions keynames to disable
        :returns: True on success, Exception otherwise

        Example::

            remove_permissions(123, ['BANDWIDTH_MANAGE'])
        """
        pretty_permissions = self.format_permission_object(permissions)
        LOGGER.warning("Removing the following permissions to %s: %s", user_id, pretty_permissions)
        return self.user_service.removeBulkPortalPermission(pretty_permissions, id=user_id)

    def permissions_from_user(self, user_id, from_user_id):
        """Sets user_id's permission to be the same as from_user_id's

        Any permissions from_user_id has will be added to user_id.
        Any permissions from_user_id doesn't have will be removed from user_id.

        :param int user_id: The user to change permissions.
        :param int from_user_id: The use to base permissions from.
        :returns: True on success, Exception otherwise.
        """

        from_permissions = self.get_user_permissions(from_user_id)
        self.add_permissions(user_id, from_permissions)
        all_permissions = self.get_all_permissions()
        remove_permissions = []

        for permission in all_permissions:
            # If permission does not exist for from_user_id add it to the list to be removed
            if _keyname_search(from_permissions, permission['keyName']):
                continue
            remove_permissions.append({'keyName': permission['keyName']})

        self.remove_permissions(user_id, remove_permissions)
        return True

    def get_user_permissions(self, user_id):
        """Returns a sorted list of a users permissions"""
        permissions = self.user_service.getPermissions(id=user_id)
        return sorted(permissions, key=itemgetter('keyName'))

    def get_logins(self, user_id, start_date=None):
        """Gets the login history for a user, default start_date is 30 days ago

        :param int id: User id to get
        :param string start_date: "%m/%d/%Y %H:%M:%s" formatted string.
        :returns: list https://softlayer.github.io/reference/datatypes/SoftLayer_User_Customer_Access_Authentication/

        Example::

            get_logins(123, '04/08/2018 0:0:0')
        """

        if start_date is None:
            date_object = datetime.datetime.today() - datetime.timedelta(days=30)
            start_date = date_object.strftime("%m/%d/%Y 0:0:0")

        date_filter = {
            'loginAttempts': {
                'createDate': {
                    'operation': 'greaterThanDate',
                    'options': [{'name': 'date', 'value': [start_date]}]
                }
            }
        }
        login_log = self.user_service.getLoginAttempts(id=user_id, filter=date_filter)
        return login_log

    def get_events(self, user_id, start_date=None):
        """Gets the event log for a specific user, default start_date is 30 days ago

        :param int id: User id to view
        :param string start_date: "%Y-%m-%dT%H:%M:%s.0000-06:00" is the full formatted string.
                                  The Timezone part has to be HH:MM, notice the : there.
        :returns: https://softlayer.github.io/reference/datatypes/SoftLayer_Event_Log/
        """

        if start_date is None:
            date_object = datetime.datetime.today() - datetime.timedelta(days=30)
            start_date = date_object.strftime("%Y-%m-%dT00:00:00")

        object_filter = {
            'userId': {
                'operation': user_id
            },
            'eventCreateDate': {
                'operation': 'greaterThanDate',
                'options': [{'name': 'date', 'value': [start_date]}]
            }
        }

        events = self.client.call('Event_Log', 'getAllObjects', filter=object_filter)
        if events is None:
            events = [{'eventName': 'No Events Found'}]
        return events

    def _get_id_from_username(self, username):
        """Looks up a username's id

        :param string username: Username to lookup
        :returns: The id that matches username.
        """
        _mask = "mask[id, username]"
        _filter = {'users': {'username': utils.query_filter(username)}}
        user = self.list_users(_mask, _filter)
        if len(user) == 1:
            return [user[0]['id']]
        elif len(user) > 1:
            raise exceptions.SoftLayerError("Multiple users found with the name: %s" % username)
        else:
            raise exceptions.SoftLayerError("Unable to find user id for %s" % username)

    def format_permission_object(self, permissions):
        """Formats a list of permission key names into something the SLAPI will respect.

        :param list permissions: A list of SLAPI permissions keyNames.
                                 keyName of ALL will return all permissions.
        :returns: list of dictionaries that can be sent to the api to add or remove permissions
        :throws SoftLayerError: If any permission is invalid this exception will be thrown.
        """
        pretty_permissions = []
        available_permissions = self.get_all_permissions()
        # pp(available_permissions)
        for permission in permissions:
            # Handle data retrieved directly from the API
            if isinstance(permission, dict):
                permission = permission['keyName']
            permission = permission.upper()
            if permission == 'ALL':
                return available_permissions
            # Search through available_permissions to make sure what the user entered was valid
            if _keyname_search(available_permissions, permission):
                pretty_permissions.append({'keyName': permission})
            else:
                raise exceptions.SoftLayerError("'%s' is not a valid permission" % permission)
        return pretty_permissions

    def gather_notifications(self, notifications_names):
        """Gets a list of notifications.

        :param list notifications_names: A list of notifications names.
        :returns: list of notifications.
        """
        notifications = []
        available_notifications = self.get_all_notifications()
        for notification in notifications_names:
            result = next((item for item in available_notifications
                           if item.get('name') == notification), None)
            if result:
                notifications.append(result)
            else:
                raise exceptions.SoftLayerError(f"{notification} is not a valid notification name")
        return notifications

    def create_user(self, user_object, password):
        """Blindly sends user_object to SoftLayer_User_Customer::createObject

        :param dictionary user_object: https://softlayer.github.io/reference/datatypes/SoftLayer_User_Customer/
        """
        LOGGER.warning("Creating User %s", user_object['username'])

        try:
            return self.user_service.createObject(user_object, password, None)
        except exceptions.SoftLayerAPIError as ex:
            if ex.faultCode == "SoftLayer_Exception_User_Customer_DelegateIamIdInvitationToPaas":
                raise exceptions.SoftLayerError("Your request for a new user was received, but it needs to be "
                                                "processed by the Platform Services API first. Barring any errors on "
                                                "the Platform Services side, your new user should be created shortly.")
            raise

    def edit_user(self, user_id, user_object):
        """Blindly sends user_object to SoftLayer_User_Customer::editObject

        :param int user_id: User to edit
        :param dictionary user_object: https://softlayer.github.io/reference/datatypes/SoftLayer_User_Customer/
        """
        return self.user_service.editObject(user_object, id=user_id)

    def add_api_authentication_key(self, user_id):
        """Calls SoftLayer_User_Customer::addApiAuthenticationKey

        :param int user_id: User to add API key to
        """
        return self.user_service.addApiAuthenticationKey(id=user_id)

    def get_api_authentication_keys(self, user_id):
        """Calls SoftLayer_User_Customer::getApiAuthenticationKeys

        :param int user_id: User to add API key to
        """
        return self.user_service.getApiAuthenticationKeys(id=user_id)

    def remove_api_authentication_key(self, user_id):
        """Calls SoftLayer_User_Customer::getApiAuthenticationKeys and

        SoftLayer_User_Customer::removeApiAuthenticationKey

        :param int user_id: User to remove API key
        """
        api_authentication_keys = self.get_api_authentication_keys(user_id)
        if len(api_authentication_keys) == 0:
            return True

        return self.user_service.removeApiAuthenticationKey(api_authentication_keys[0]['id'])

    def vpn_manual(self, user_id, value):
        """Enable or disable the manual config of subnets.

        :param int user_id: User to edit.
        :param bool value: Value for vpnManualConfig flag.
        """
        user_object = {'vpnManualConfig': value}
        return self.edit_user(user_id, user_object)

    def vpn_enable_or_disable(self, user_id, value):
        """Enable or Disable vpn for a user.

        :param int user_id: User to edit.
        :param bool value: Value for vpn enable flag.
        :param bool value: Value for vpn disable flag.
        """
        user_object = {'sslVpnAllowedFlag': value}
        return self.edit_user(user_id, user_object)

    def vpn_subnet_add(self, user_id, subnet_ids):
        """Add subnets for a user.

        :param int user_id: User to edit.
        :param list subnet_ids: list of subnet Ids.
        """
        overrides = [{"userId": user_id, "subnetId": subnet_id} for subnet_id in subnet_ids]
        return_value = self.override_service.createObjects(overrides)
        update_success = self.user_service.updateVpnUser(id=user_id)
        if not update_success:
            raise exceptions.SoftLayerAPIError("Overrides created, but unable to update VPN user")
        return return_value

    def vpn_subnet_remove(self, user_id, subnet_ids):
        """Remove subnets for a user.

        :param int user_id: User to edit.
        :param list subnet_ids: list of subnet Ids.
        """
        overrides = self.get_overrides_list(user_id, subnet_ids)
        return_value = self.override_service.deleteObjects(overrides)
        update_success = self.user_service.updateVpnUser(id=user_id)
        if not update_success:
            raise exceptions.SoftLayerAPIError("Overrides deleted, but unable to update VPN user")
        return return_value

    def get_overrides_list(self, user_id, subnet_ids):
        """Converts a list of subnets to a list of overrides.

        :param int user_id: The ID of the user.
        :param list subnet_ids: A list of subnets.
        :returns: A list of overrides associated with the given subnets.
        """

        overrides_list = []
        matching_overrides = {}
        output_error = "Subnet {} does not exist in the subnets assigned for user {}"
        _mask = 'mask[id,subnetId]'
        overrides = self.user_service.getOverrides(id=user_id, mask=_mask)
        for subnet in subnet_ids:
            for override in overrides:
                if int(subnet) == override.get('subnetId'):
                    matching_overrides = override
                    break
            if matching_overrides.get('subnetId') is None:
                raise exceptions.SoftLayerError(output_error.format(subnet, user_id))

            overrides_list.append(matching_overrides)

        return overrides_list

    def grant_hardware_access(self, user_id, hardware_id):
        """Grants the user access to a single hardware device.

        :param int user_id: User Id
        :param int hardware_id: Hardware Id
        :returns: true
        """
        return self.user_service.addHardwareAccess(hardware_id, id=user_id)

    def grant_virtual_access(self, user_id, virtual_id):
        """Grants the user access to a single VS device.

        :param int user_id: User Id
        :param int virtual_id: Hardware Id
        :returns: true
        """
        return self.user_service.addVirtualGuestAccess(virtual_id, id=user_id)

    def grant_dedicated_access(self, user_id, dedicated_id):
        """Grants the user access to a single dedicated host device.

        :param int user_id: User Id
        :param int dedicated_id: Dedicatd Host Id
        :returns: true
        """
        return self.user_service.addDedicatedHostAccess(dedicated_id, id=user_id)

    def remove_hardware_access(self, user_id, hardware_id):
        """Remove hardware from a portal user’s hardware access list.

        :param int user_id: User Id
        :param int hardware_id: Hardware Id
        :returns: true
        """
        return self.user_service.removeHardwareAccess(hardware_id, id=user_id)

    def remove_virtual_access(self, user_id, virtual_id):
        """Remove hardware from a portal user’s virtual guests access list.

        :param int user_id: User Id
        :param int hardware_id: Hardware Id
        :returns: true
        """
        return self.user_service.removeVirtualGuestAccess(virtual_id, id=user_id)

    def remove_dedicated_access(self, user_id, dedicated_id):
        """Remove hardware from a portal user’s dedicated host access list.

        :param int user_id: User Id
        :param int dedicated_id: Dedicated Host Id
        :returns: true
        """
        return self.user_service.removeDedicatedHostAccess(dedicated_id, id=user_id)

    def get_user_hardware(self, user_id):
        """User Hardware list.

        :param int user_id: User Id
        :return: List hardware relate to user
        """
        return self.user_service.getHardware(id=user_id)

    def get_user_dedicated_host(self, user_id):
        """User dedicate host list.

        :param int user_id: User Id
        :return: List dedicated host relate to user
        """
        return self.user_service.getDedicatedHosts(id=user_id)

    def get_user_virtuals(self, user_id):
        """User virtual guest list.

        :param int user_id: User Id
        :return: List virtual guest relate to user
        """
        return self.user_service.getVirtualGuests(id=user_id)

    def update_vpn_password(self, user_id, password):
        """Update a user's VPN password.

        :param int user_id: User Id
        :param string password: Password
        :returns: true
        """
        return self.user_service.updateVpnPassword(password, id=user_id)


def _keyname_search(haystack, needle):
    for item in haystack:
        if item.get('keyName') == needle:
            return True
    return False
